# -*- coding: utf-8 -*-
"""Featurizer that runs RASPA to calculate energy grids."""
import os
from glob import glob
from typing import List, Tuple, Union

import numpy as np
import pandas as pd
from pymatgen.core import IStructure, Structure

from mofdscribe.featurizers.base import MOFBaseFeaturizer
from mofdscribe.featurizers.utils.histogram import get_rdf
from mofdscribe.featurizers.utils.raspa.resize_uc import resize_unit_cell
from mofdscribe.featurizers.utils.raspa.run_raspa import detect_raspa_dir, run_raspa

from ..utils.extend import operates_on_istructure, operates_on_structure

__all__ = ["EnergyGridHistogram"]
GRID_INPUT_TEMPLATE = """SimulationType  MakeASCIGrid

Forcefield      Local

Framework 0
FrameworkName input
UnitCells {unit_cells}

CutOff                        {cutoff}

NumberOfGrids {num_grids}
GridTypes {grid_types}
SpacingVDWGrid {vdw_spacing}
"""


def parse_energy_grids(directory: Union[str, os.PathLike]) -> dict:
    grids = glob(os.path.join(directory, "ASCI_Grids", "*.grid"))
    energies = {}
    for grid in grids:
        name = os.path.basename(grid).split(".")[0].replace("asci_grid_", "")
        energies[name] = read_ascii_grid(grid)
    return energies


def read_ascii_grid(filename: Union[str, os.PathLike]) -> pd.DataFrame:
    """Read an ASCII grid file into a pandas DataFrame.

    Args:
        filename (Union[str, os.PathLike]): The path to the file to read.

    Returns:
        A pandas DataFrame containing the grid data.
    """
    df = pd.read_csv(
        filename,
        delim_whitespace=True,
        header=None,
        names=["x", "y", "z", "energy", "deriv_x", "deriv_y", "deriv_z"],
        na_values="?",
    )
    df = df.astype(np.float)
    return df


@operates_on_istructure
@operates_on_structure
class EnergyGridHistogram(MOFBaseFeaturizer):
    """Computes the energy grid histograms as originally proposed by Bucior et al. [Bucior2019]_.

    Conventionally, energy grids can be used to speed up molecular simulations.
    The idea is that the interactions between the guest and host are
    pre-computed on a fine grid and then only need to be looked up (instead of
    re-computed all the time).
    Bucior et al. proposed (effectively) a dimensionality reduction of the
    energy grid by making a histogram of the energies. For H2 they used bins of
    a width of 1 kJ mol−1 ranging from −10 kJ mol−1 (attractive) to 0 kJ mol−1.
    For methane, they used bins in 2 kJ mol−1 increments between −26 and 0 kJ
    mol−1, again with a repulsion bin.

    This approach has also been used, for example, `Li et al. (2021)
    <https://aip.scitation.org/doi/10.1063/5.0050823>`_ and [Bucior2021]_.

    References:
        .. [Bucior2021]  Li, Z.; Bucior, B. J.; Chen, H.; Haranczyk, M.; Siepmann, J. I.;
            Snurr, R. Q. Machine Learning Using Host/Guest Energy Histograms to
            Predict Adsorption in Metal–Organic Frameworks: Application to Short
            Alkanes and Xe/Kr Mixtures. J. Chem. Phys. 2021, 155 (1), 014701.
            https://doi.org/10.1063/5.0050823.
    """

    def __init__(
        self,
        raspa_dir: Union[str, os.PathLike, None] = None,
        grid_spacing: float = 1.0,
        bin_size_vdw: float = 1,
        min_energy_vdw: float = -40,
        max_energy_vdw: float = 0,
        cutoff: float = 12,
        mof_ff: str = "UFF",
        mol_ff: str = "TraPPE",
        mol_name: str = "CO2",
        sites: Tuple[str] = ("C_co2",),
        tail_corrections: bool = True,
        mixing_rule: str = "Lorentz-Berthelot",
        shifted: bool = False,
        separate_interactions: bool = True,
        run_eqeq: bool = True,
        primitive: bool = False,
    ):
        """Construct the EnergyGridHistogram class.

        Args:
            raspa_dir (Union[str, Path, None]): Path to the raspa
                directory (with lib, bin, share) subdirectories.
                If `None` we will look for the `RASPA_DIR` environment variable.
                Defaults to None.
            grid_spacing (float): Spacing for the energy grids.
                Bucior et al. (2018) used 1.0 A. Defaults to 1.0.
            bin_size_vdw (float): Size of bins for the energy
                histogram. Defaults to 1.
            min_energy_vdw (float): Minimum
                energy for the energy histogram (defining start of first bin).
                Defaults to -10.
            max_energy_vdw (float): Maximum energy for energy
                histogram (defining last bin).
                Defaults to 0.
            cutoff (float): Cutoff for the Van-der-Waals interaction.
                Defaults to 12.
            mof_ff (str): Name of the forcefield used
                for the framework. Defaults to "UFF".
            mol_ff (str): Name of the forcefield used for the guest molecule.
                Defaults to "TraPPE".
            mol_name (str): Name of the guest molecule.
                Defaults to "CO2".
            sites (Tuple[str]): Name of the Van-der-Waals sites
                for which the energy diagrams are computed.
                Defaults to ("C_co2",).
            tail_corrections (bool): If true, use analytical
                tail-correction for the contribution of the interaction potential after the
                cutoff. Defaults to True.
            mixing_rule (str): Mixing rule for framework and guest
                molecule force field. Available options are `Jorgenson` and
                `Lorentz-Berthelot`. Defaults to "Lorentz-Berthelot".
            shifted (bool): If true, shifts the potential to equal to zero at the
                cutoff. Defaults to False.
            separate_interactions (bool): If True use framework's force field
                for framework-molecule interactions.
                Defaults to True.
            run_eqeq (bool): If true, runs EqEq to compute charges.
                Defaults to True.
            primitive (bool): If True, the structure is reduced to its primitive
                form before the descriptor is computed. Defaults to True.

        Raises:
            ValueError: If the `raspa_dir` is not a valid directory.
        """
        self.raspa_dir = raspa_dir if raspa_dir else os.environ.get("RASPA_DIR")
        if self.raspa_dir is None:
            try:
                self.raspa_dir = detect_raspa_dir()
            except ValueError:
                raise ValueError(
                    "Please set the RASPA_DIR environment variable or provide the path for the class initialization."
                )
        self.grid_spacing = grid_spacing
        self.bin_size_vdw = bin_size_vdw
        self.min_energy_vdw = min_energy_vdw
        self.max_energy_vdw = max_energy_vdw
        self.cutoff = cutoff
        self.mof_ff = mof_ff
        self.mol_ff = mol_ff
        self.mol_name = mol_name
        self.sites = sites
        self.tail_corrections = tail_corrections
        self.mixing_rule = mixing_rule
        self.shifted = shifted
        self.separate_interactions = separate_interactions
        self.run_eqeq = run_eqeq
        super().__init__(primitive=primitive)

    def fit_transform(self, structures: List[Union[Structure, IStructure]]):
        ...

    def fit(self, structure: Union[Structure, IStructure]):
        ...

    def _get_grid(self):
        return np.arange(self.min_energy_vdw, self.max_energy_vdw, self.bin_size_vdw)

    def feature_labels(self) -> List[str]:
        grid = self._get_grid()
        labels = []
        for site in self.sites:
            for grid_point in grid:
                labels.append(f"energygridhist_{self.mol_name}_{site}_{grid_point}")
        return labels

    def _featurize(self, s: Union[Structure, IStructure]) -> np.array:
        ff_molecules = {self.mol_name: self.mol_ff}

        parameters = {
            "ff_framework": self.mof_ff,
            "ff_molecules": ff_molecules,
            "shifted": self.shifted,
            "tail_corrections": self.tail_corrections,
            "mixing_rule": self.mixing_rule,
            "separate_interactions": self.separate_interactions,
        }
        replicas = resize_unit_cell(s, self.cutoff)
        ucells = f"{replicas[0]} {replicas[1]} {replicas[2]}"

        simulation_script = GRID_INPUT_TEMPLATE.format(
            unit_cells=ucells,
            cutoff=self.cutoff,
            num_grids=len(self.sites),
            grid_types=" ".join(self.sites),
            vdw_spacing=self.grid_spacing,
        )
        res = run_raspa(
            s, self.raspa_dir, simulation_script, parameters, parse_energy_grids, self.run_eqeq
        )
        output = []
        for _, v in res.items():
            output.append(
                get_rdf(
                    v["energy"].values,
                    self.min_energy_vdw,
                    self.max_energy_vdw,
                    self.bin_size_vdw,
                    None,
                    None,
                    normalized=False,
                )
            )
        return np.concatenate(output)

    def citations(self) -> List[str]:
        return [
            "@article{Bucior2019,"
            "doi = {10.1039/c8me00050f},"
            "url = {https://doi.org/10.1039/c8me00050f},"
            "year = {2019},"
            "publisher = {Royal Society of Chemistry ({RSC})},"
            "volume = {4},"
            "number = {1},"
            "pages = {162--174},"
            "author = {Benjamin J. Bucior and N. Scott Bobbitt and Timur Islamoglu "
            "and Subhadip Goswami and Arun Gopalan and Taner Yildirim and "
            "Omar K. Farha and Neda Bagheri and Randall Q. Snurr},"
            "title = {Energy-based descriptors to rapidly predict hydrogen storage "
            "in metal{\textendash}organic frameworks},"
            r"journal = {Molecular Systems Design {\&}amp$\mathsemicolon$ Engineering}"
            "}",
            "@article{Dubbeldam2015,"
            "doi = {10.1080/08927022.2015.1010082},"
            "url = {https://doi.org/10.1080/08927022.2015.1010082},"
            "year = {2015},"
            "month = feb,"
            "publisher = {Informa {UK} Limited},"
            "volume = {42},"
            "number = {2},"
            "pages = {81--101},"
            r"author = {David Dubbeldam and Sof{'{\i}}a Calero and Donald E. Ellis "
            "and Randall Q. Snurr},"
            "title = {{RASPA}: molecular simulation software for adsorption and "
            "diffusion in flexible nanoporous materials},"
            "journal = {Molecular Simulation}"
            "}",
        ]

    def implementors(self):
        return ["Kevin Maik Jablonka", "David Dubbeldam and RASPA authors"]



# -*- coding: utf-8 -*-
"""Guest-centered atomic-property weighted autocorrelation function."""
from functools import cached_property
from typing import List, Optional, Tuple, Union

import numpy as np
from element_coder import encode
from matminer.featurizers.base import BaseFeaturizer
from pymatgen.core import IStructure, Structure

from mofdscribe.featurizers.hostguest.utils import HostGuest, _extract_host_guest

from ..utils.aggregators import AGGREGATORS

__all__ = ["GuestCenteredAPRDF"]


class GuestCenteredAPRDF(BaseFeaturizer):
    """Guest-centered atomic-property weighted autocorrelation function.

    This is a modification of the AP-RDF that is centered on the guest atoms.
    That is, we only consider the distances between the guest atoms and the host atoms (within the cutoffs).

    For more details about the AP-RDF, see `mofdscribe.featurizers.chemistry.aprdf.APRDF`.
    """

    def __init__(
        self,
        cutoff: float = 20.0,
        lower_lim: float = 2,
        bin_size: float = 0.25,
        b_smear: Union[float, None] = 10,
        properties: Tuple[str, int] = ("X", "electron_affinity"),
        aggregations: Tuple[str] = ("avg", "product", "diff"),
        local_env_method: str = "vesta",
        normalize: bool = False,
    ):
        """Set up an atomic property (AP) weighted radial distribution function.

        Args:
            cutoff (float): Consider neighbors up to this value (in
                Angstrom). Defaults to 20.0.
            lower_lim (float): Lowest distance (in Angstrom) to consider.
                Defaults to 2.
            bin_size (float): Bin size for binning.
                Defaults to 0.25.
            b_smear (Union[float, None]): Band width for Gaussian smearing.
                If None, the unsmeared histogram is used. Defaults to 10.
            properties (Tuple[str, int]): Properties used for calculation of the AP-RDF.
                All properties of `pymatgen.core.Species` are available in
                addition to the integer `1` that will set P_i=P_j=1. Defaults to
                ("X", "electron_affinity").
            aggregations (Tuple[str]): Methods used to combine the
                properties.
                See `mofdscribe.featurizers.utils.aggregators.AGGREGATORS` for available
                options. Defaults to ("avg", "product", "diff").
            local_env_method (str): Method used to compute the structure graph.
            normalize (bool): If True, the histogram is normalized by dividing
                by the number of atoms. Defaults to False.
        """
        self.lower_lim = lower_lim
        self.cutoff = cutoff
        self.bin_size = bin_size
        self.properties = properties
        self._local_env_method = local_env_method
        self.b_smear = b_smear
        self.aggregations = aggregations
        self.normalize = normalize

    @cached_property
    def _bins(self):
        num_bins = int((self.cutoff - self.lower_lim) // self.bin_size)
        bins = np.linspace(self.lower_lim, self.cutoff, num_bins)
        return bins

    def _get_feature_labels(self):
        aprdfs = np.empty(
            (len(self.properties), len(self.aggregations), len(self._bins)), dtype=object
        )
        for pi, prop in enumerate(self.properties):
            for ai, aggregation in enumerate(self.aggregations):
                for bin_index, _ in enumerate(self._bins):
                    aprdfs[pi][ai][bin_index] = f"guest_aprdf_{prop}_{aggregation}_{bin_index}"

        return list(aprdfs.flatten())

    def _extract_host_guest(
        self,
        structure: Optional[Union[Structure, IStructure]] = None,
        host_guest: Optional[HostGuest] = None,
    ):
        return _extract_host_guest(
            structure=structure,
            host_guest=host_guest,
            remove_guests=True,
            operates_on="molecule",
            local_env_method=self._local_env_method,
        )

    def featurize(
        self,
        structure: Optional[Union[Structure, IStructure]],
        host_guest: Optional[HostGuest] = None,
    ) -> np.ndarray:
        """
        Compute the features of the host and the guests and aggregate them.

        Args:
            structure (Optional[Union[Structure, IStructure]]): The structure to featurize.
            host_guest (Optional[HostGuest]): The host_guest to featurize.
                If you provide this, you must not provide structure.

        Returns:
            np.ndarray: The features of the host and the guests.

        Raises:
            ValueError: If we cannot detect a host.
        """
        host_guest = self._extract_host_guest(structure=structure, host_guest=host_guest)

        if not host_guest.host:
            raise ValueError(
                "Did not find a framework. This is required for the host guest featurizer."
            )

        if not host_guest.guests:
            raise ValueError(
                "Did not find any guests. This is required for the host guest featurizer."
            )

        flattened_guests = []
        for guest in host_guest.guests:
            flattened_guests.extend(guest.sites)

        bins = self._bins
        aprdfs = np.zeros((len(self.properties), len(self.aggregations), len(bins)))

        # todo: use numba to speed up
        for _i, guest in enumerate(flattened_guests):
            guest_frac_coords = host_guest.host.lattice.get_fractional_coords(guest.coords)
            for _j, site in enumerate(host_guest.host):
                dist, _ = host_guest.host.lattice.get_distance_and_image(
                    guest_frac_coords, site.frac_coords
                )
                if dist < self.cutoff and dist > self.lower_lim:
                    bin_idx = int((dist - self.lower_lim) // self.bin_size)
                    for pi, prop in enumerate(self.properties):
                        for ai, agg in enumerate(self.aggregations):
                            p0 = encode(guest.specie, prop)
                            p1 = encode(site.specie, prop)

                            agg_func = AGGREGATORS[agg]
                            p = agg_func([p0, p1])
                            aprdfs[pi][ai][bin_idx] += p * np.exp(
                                -self.b_smear * (dist - bins[bin_idx]) ** 2
                            )

        if self.normalize:
            aprdfs /= len(host_guest.guests)

        return aprdfs.flatten()

    def feature_labels(self) -> List[str]:
        return self._get_feature_labels()

    def citations(self) -> List[str]:
        return [
            "@article{Fernandez2013,"
            "doi = {10.1021/jp404287t},"
            "url = {https://doi.org/10.1021/jp404287t},"
            "year = {2013},"
            "month = jul,"
            "publisher = {American Chemical Society ({ACS})},"
            "volume = {117},"
            "number = {27},"
            "pages = {14095--14105},"
            "author = {Michael Fernandez and Nicholas R. Trefiak and Tom K. Woo},"
            "title = {Atomic Property Weighted Radial Distribution Functions "
            "Descriptors of Metal{\textendash}Organic Frameworks for the Prediction "
            "of Gas Uptake Capacity},"
            "journal = {The Journal of Physical Chemistry C}"
            "}"
        ]

    def implementors(self):
        return ["Kevin Maik Jablonka"]



# -*- coding: utf-8 -*-
from .arabg_dataset import ARABGDataset  # noqa: F401
from .bw_dataset import BWDataset  # noqa: F401
from .core_dataset import CoREDataset  # noqa: F401
from .qmof_dataset import QMOFDataset  # noqa: F401



# -*- coding: utf-8 -*-
"""Featurizer that computes 3D voxelgrids."""
from typing import List, Tuple, Union

import numpy as np
from element_coder import encode
from pymatgen.core import Element, IStructure, Structure

from mofdscribe.featurizers.base import MOFBaseFeaturizer

from ._voxelgrid import VoxelGrid as VGBase


def make_supercell(
    coords: np.ndarray,
    elements: List[int],
    lattice: Tuple[float, float, float],
    size: float,
    min_size: float = -5,
) -> np.ndarray:
    """Generate cubic supercell of a given size.

    Args:
        coords (np.ndarray): matrix of xyz coordinates of the system
        elements (List[int]): atomic numbers of every site
        lattice (Tuple[float, float, float]): lattice constants of the system
        size (float): dimension size of cubic cell, e.g., 10x10x10
        min_size (float): minimum axes size to keep negative xyz
            coordinates from the original cell. Defaults to -5.

    Returns:
        new_cell: supercell array
    """
    a, b, c = lattice
    elements = np.array(elements).reshape(-1, 1)
    xyz_periodic_copies = [coords]
    element_periodic_copies = [elements]

    min_range = -3  # we aren't going in the minimum direction too much, so can make this small
    max_range = 20  # make this large enough, but can modify if wanting an even larger cell

    for x in range(-min_range, max_range):
        for y in range(0, max_range):
            for z in range(0, max_range):
                if x == y == z == 0:
                    continue
                add_vector = x * a + y * b + z * c
                xyz_periodic_copies.append(coords + add_vector)
                element_periodic_copies.append(elements)

    # Combine into one array
    xyz_periodic_total = np.vstack(xyz_periodic_copies)
    elements_all = np.vstack(element_periodic_copies)

    # Filter out all atoms outside of the cubic box
    filter_mask_a = np.max(xyz_periodic_total[:, :3], axis=1) < size
    new_cell = xyz_periodic_total[filter_mask_a]

    filter_mask_b = np.min(new_cell[:, :3], axis=1) > min_size
    new_cell = new_cell[filter_mask_b]

    elements_all = elements_all[filter_mask_a]
    elements_all = elements_all[filter_mask_b]

    return new_cell, elements_all.flatten()


def compute_properties(numbers: np.array, properties: Tuple[str]) -> np.array:
    property_lookup = {}

    unique_numbers = np.unique(numbers)

    for number in unique_numbers:
        element = Element.from_Z(number)
        prop_vec = []
        for prop in properties:
            prop_vec.append(encode(element.symbol, prop))

        property_lookup[number] = prop_vec

    property_array = []
    for number in numbers:
        property_array.append(property_lookup[number])

    return np.array(property_array)


# ToDo: Potentially, we could also do the substruture-based approach for chemistry.
# Not sure though how useful it would be as it would lead to quite large feature vectors.
class VoxelGrid(MOFBaseFeaturizer):
    """
    Describe the structure using a voxel grid.

    For this, we first compute a supercell, the "voxelize" the point cloud.

    For setting the value of the voxels, different options are available:
    Geometry Aggregations: - `binary`: 1 if the voxel is occupied, 0 otherwise -
    `density`: the number of atoms in the voxel / total number of atoms - `TDF`:
    truncated distance function. Value between 0 and 1 indicating the distance
    between the voxel's center and the closest point. 1 on the surface, 0 on
    voxels further than 2 * voxel side.

    Properties: Alternatively/additionally one can use the average of any
    available properties of pymatgen Element objects.
    """

    def __init__(
        self,
        min_size: float = 30,
        n_x: int = 25,
        n_y: int = 25,
        n_z: int = 25,
        geometry_aggregations: Tuple[str] = ("binary",),
        properties: Tuple[str, int] = ("X", "electron_affinity"),
        flatten: bool = True,
        regular_bounding_box: bool = True,
        primitive: bool = False,
    ):
        """Initialize a VoxelGrid featurizer.

        Args:
            min_size (float): Minimum supercell size in Angstrom.
                Defaults to 30.
            n_x (int): Number of bins in x direction
                (Hung et al used 30 and 60 at a cell size of 60). Defaults to 25.
            n_y (int): Number of bins in x direction
                (Hung et al used 30 and 60 at a cell size of 60). Defaults to 25.
            n_z (int): Number of bins in x direction
                (Hung et al used 30 and 60 at a cell size of 60). Defaults to 25.
            geometry_aggregations (Union[Tuple["density" | "binary" | "TDF"], None]):
                Mode for encoding the occupation of voxels.
                * binary: 0 for empty voxels, 1 for occupied.
                * density: number of points inside voxel / total number of points.
                * TDF: Truncated Distance Function. Value between 0 and 1 indicating the distance,
                between the voxel's center and the closest point. 1 on the surface,
                0 on voxels further than 2 * voxel side.
                Defaults to ("binary",).
            properties (Union[Tuple[str, int], None]): Properties used for calculation of the AP-RDF.
                All properties of `pymatgen.core.Species` are available. Defaults to ("X", "electron_affinity").
            flatten (bool): It true, flatten the 3D voxelgrid to 1D array. Defaults to True.
            regular_bounding_box (bool): If True, the bounding box of the point cloud will be adjusted
                in order to have all the dimensions of equal length.
                Defaults to True.
            primitive (bool): If True, the structure is reduced to its primitive
                form before the descriptor is computed. Defaults to True.
        """
        self.min_size = min_size
        self.n_x = n_x
        self.n_y = n_y
        self.n_z = n_z
        self._num_voxels = n_x * n_y * n_z
        self.properties = properties
        self.geometry_aggregations = geometry_aggregations
        self.flatten = flatten
        self.regular_bounding_box = regular_bounding_box
        super().__init__(primitive=primitive)

    def _featurize(self, structure: Union[IStructure, Structure]) -> np.ndarray:
        coords, numbers = make_supercell(
            structure.cart_coords, structure.atomic_numbers, structure.lattice.matrix, self.min_size
        )

        if self.properties is not None:
            properties = compute_properties(numbers, self.properties)
        else:
            properties = None

        vg = VGBase(
            coords,
            properties,
            self.n_x,
            self.n_y,
            self.n_z,
            regular_bounding_box=self.regular_bounding_box,
        )
        vg.compute()
        features = []
        if self.geometry_aggregations is not None:
            for agg in self.geometry_aggregations:
                features.append(
                    vg.get_feature_vector(
                        agg,
                        flatten=self.flatten,
                    )
                )

        if self.properties is not None:
            for i, _ in enumerate(self.properties):
                features.append(
                    vg.get_feature_vector(
                        i,
                        flatten=self.flatten,
                    )
                )

        if self.flatten:
            return np.stack(features).flatten()
        return features

    def _get_feature_labels(self):
        feature_labels = []
        for geometry_aggregation in self.geometry_aggregations:
            for voxel in range(self._num_voxels):
                feature_labels.append(f"voxelgrid_{geometry_aggregation}_{voxel}")
        for prop in self.properties:
            for voxel in range(self._num_voxels):
                feature_labels.append(f"voxelgrid_{prop}_{voxel}")

        return feature_labels

    def feature_labels(self) -> List[str]:
        return self._get_feature_labels()

    def citations(self) -> List[str]:
        return [
            "@article{Hung2022,"
            "doi = {10.1021/acs.jpcc.1c09649},"
            "url = {https://doi.org/10.1021/acs.jpcc.1c09649},"
            "year = {2022},"
            "month = jan,"
            "publisher = {American Chemical Society ({ACS})},"
            "volume = {126},"
            "number = {5},"
            "pages = {2813--2822},"
            "author = {Ting-Hsiang Hung and Zhi-Xun Xu and Dun-Yen Kang "
            "and Li-Chiang Lin},"
            "title = {Chemistry-Encoded Convolutional Neural Networks for "
            "Predicting Gaseous Adsorption in Porous Materials},"
            "journal = {The Journal of Physical Chemistry C}"
            "}",
            "@article{Cho2021,"
            "doi = {10.1021/acs.jpclett.1c00293},"
            "url = {https://doi.org/10.1021/acs.jpclett.1c00293},"
            "year = {2021},"
            "month = mar,"
            "publisher = {American Chemical Society ({ACS})},"
            "volume = {12},"
            "number = {9},"
            "pages = {2279--2285},"
            "author = {Eun Hyun Cho and Li-Chiang Lin},"
            "title = {Nanoporous Material Recognition via 3D Convolutional "
            "Neural Networks: Prediction of Adsorption Properties},"
            "journal = {The Journal of Physical Chemistry Letters}"
            "}",
        ]

    def implementors(self) -> List[str]:
        return ["Kevin Maik Jablonka"]



# -*- coding: utf-8 -*-
import pytest

from mofdscribe.featurizers.chemistry.price import PriceLowerBound


def test_price_lower_bound(hkust_structure):
    """Comparing with the original implementation."""
    pricer = PriceLowerBound()
    feats = pricer.featurize(hkust_structure)
    assert len(feats) == 2
    assert feats[0] == pytest.approx(4.176635436396251)
    assert feats[1] == pytest.approx(3.671662426852288)

    pricer = PriceLowerBound(("per_atom",))
    feats = pricer.featurize(hkust_structure)
    assert len(feats) == 1
    assert feats[0] == pytest.approx(64.77758513112447)



# -*- coding: utf-8 -*-
"""Test the AMD featurizer."""
import os

import amd
import pytest

from mofdscribe.featurizers.chemistry.amd import AMD

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


def test_amd_consistency(hkust_structure):
    """Test that AMD is consistent with the original package."""
    amd_mofdscribe = AMD(k=100, atom_types=None)
    amd_hkust_mofdscribe = amd_mofdscribe.featurize(hkust_structure)

    reader = amd.CifReader(os.path.join(THIS_DIR, "..", "..", "test_files", "HKUST-1.cif"))
    amds = [amd.AMD(crystal, 100) for crystal in reader]
    for feat, amd_feat in zip(amds[0], amd_hkust_mofdscribe):
        assert feat == pytest.approx(amd_feat, abs=1e-3)

    amd_mofdscribe = AMD(k=100, aggregations=("mean", "std"))
    amd_hkust_mofdscribe = amd_mofdscribe.featurize(hkust_structure)
    assert len(amd_hkust_mofdscribe) == 100 * 2 * 4
    assert len(amd_mofdscribe.feature_labels()) == 100 * 2 * 4



import pandas as pd

from mofdscribe.datasets.structuredataset import FrameDataset, StructureDataset


def test_structuredataset(dataset_files, dataset_folder):
    # make a dataset only from the files
    structures, frame = dataset_files
    ds = StructureDataset(structures)
    assert len(ds) == len(structures)
    hashes = ds.get_decorated_graph_hashes([0, 1, 2, 3])
    assert len(hashes) == 4
    densities = ds.get_densities([0, 1, 2, 3])
    assert len(densities) == 4

    # make a dataset from the files and a dataframe
    frame = pd.read_json(frame[0])
    ds = StructureDataset(
        structures,
        frame,
        structure_name_column="info.basename",
        decorated_graph_hash_column="info.decorated_graph_hash",
    )
    # only two of them are in the dataframe
    assert len(ds) == 2
    hashes = ds.get_decorated_graph_hashes([0, 1])
    assert len(hashes) == 2

    # make a dataset from a folder and a dataframe
    ds = StructureDataset.from_folder_and_dataframe(
        dataset_folder,
        dataframe=frame,
        structure_name_column="info.basename",
        decorated_graph_hash_column="info.decorated_graph_hash",
    )
    # only two of them are in the dataframe
    assert len(ds) == 2
    hashes = ds.get_decorated_graph_hashes([0, 1])
    assert len(hashes) == 2


def test_framedataset(dataset_files):
    _, frame = dataset_files
    frame = pd.read_json(frame[0])
    ds = FrameDataset(
        frame,
        structure_name_column="info.basename",
        decorated_graph_hash_column="info.decorated_graph_hash",
    )
    # only two of them are in the dataframe
    assert len(ds) == 2
    hashes = ds.get_decorated_graph_hashes([0, 1])
    assert len(hashes) == 2



# -*- coding: utf-8 -*-
"""Test the geometric properties featurizers."""

from pytest import approx

from mofdscribe.featurizers.pore.geometric_properties import (
    AccessibleVolume,
    PoreDiameters,
    PoreSizeDistribution,
    RayTracingHistogram,
    SurfaceArea,
    _parse_res_zeopp,
    _parse_sa_zeopp,
    _parse_volpo_zeopp,
)

from ..helpers import is_jsonable

SA_SAMPLE_OUTPUT = """@ EDI.sa Unitcell_volume: 307.484   Density: 1.62239

ASA_A^2: 60.7713 ASA_m^2/cm^3: 1976.4 ASA_m^2/g: 1218.21

NASA_A^2: 0 NASA_m^2/cm^3: 0 NASA_m^2/g: 0"""

VOLPO_SAMPLE_OUTPUT = """@ EDI.vol Unitcell_volume: 307.484   Density: 1.62239

AV_A^3: 22.6493 AV_Volume_fraction: 0.07366 AV_cm^3/g: 0.0454022

NAV_A^3: 0 NAV_Volume_fraction: 0 NAV_cm^3/g: 0"""


RES_SAMPLE_OUTPUT = "output_file.res    1.70107 0.95106  1.64805"


def test_parse_sa():
    """Ensure that the parser works as expected."""
    parsed = _parse_sa_zeopp(SA_SAMPLE_OUTPUT)
    assert parsed["unitcell_volume"] == approx(307.484, 0.1)
    assert parsed["density"] == approx(1.62239, 0.1)
    assert parsed["asa_a2"] == approx(60.7713, 0.1)
    assert parsed["asa_m2cm3"] == approx(1976.4, 0.1)
    assert parsed["asa_m2g"] == approx(1218.21, 0.1)
    assert parsed["nasa_a2"] == 0
    assert parsed["nasa_m2cm3"] == 0
    assert parsed["nasa_m2g"] == 0


def test_parse_res():
    """Ensure that the parser works as expected."""
    res = _parse_res_zeopp(RES_SAMPLE_OUTPUT)
    assert res == {
        "lis": 1.70107,  # largest included sphere
        "lifs": 0.95106,  # largest free sphere
        "lifsp": 1.64805,  # largest included sphere along free sphere path
    }


def test_parse_volpo_zeopp():
    """Ensure that the parser works as expected."""
    res = _parse_volpo_zeopp(VOLPO_SAMPLE_OUTPUT)
    assert res["unitcell_volume"] == approx(307.484, 0.1)
    assert res["density"] == approx(1.62239, 0.1)
    # assert res["av_a3"] == approx(22.6493, 0.1)
    assert res["av_volume_fraction"] == approx(0.07366, 0.1)
    assert res["av_cm3g"] == approx(0.0454022, 0.1)
    assert res["nav_a3"] == 0
    assert res["nav_volume_fraction"] == 0
    assert res["nav_cm3g"] == 0


def test_pore_diameters(hkust_structure):
    """Ensure that the featurizer works as expected."""
    pd = PoreDiameters()
    assert pd.feature_labels() == ["lis", "lifs", "lifsp"]
    result = pd.featurize(hkust_structure)
    assert len(result) == 3
    assert result[0] == approx(13.21425, 0.1)
    assert result[1] == approx(6.66829, 0.1)
    assert result[2] == approx(13.21425, 0.1)
    assert is_jsonable(dict(zip(pd.feature_labels(), result)))
    assert result.ndim == 1


def test_surface_area(hkust_structure):
    """Ensure that the featurizer works as expected."""
    sa = SurfaceArea()
    expected_labels = [
        "uc_volume",
        "density",
        "asa_a2",
        "asa_m2cm3",
        "asa_m2g",
        "nasa_a2",
        "nasa_m2cm3",
        "nasa_m2g",
    ]

    for el, fl in zip(expected_labels, sa.feature_labels()):
        assert el in fl

    result = sa.featurize(hkust_structure)
    assert len(result) == 8

    assert result[0] == approx(4570.21, 0.1)
    assert result[1] == approx(8.79097e-01, 0.1)
    # assert result[2] == approx(5.13510e03, 0.1)
    assert result[3] == approx(2.80901e03, 0.1)
    assert result[4] == approx(3.19533e03, 0.1)
    assert result[5] == 0
    assert result[6] == 0
    assert result[7] == 0
    assert is_jsonable(dict(zip(sa.feature_labels(), result)))
    assert result.ndim == 1


def test_accessible_volume(hkust_structure):
    """Ensure that the featurizer works as expected."""
    av = AccessibleVolume()
    expected_labels = [
        "uc_volume",
        "density",
        "av_a2",
        "av_volume_fraction",
        "av_cm3g",
        "nav_a3",
        "nav_volume_fraction",
        "nav_cm3g",
    ]

    for el, fl in zip(expected_labels, av.feature_labels()):
        assert el in fl

    result = av.featurize(hkust_structure)
    assert len(result) == 8
    assert result[0] == approx(4570.21, 0.1)
    assert result[1] == approx(8.79097e-01, 0.1)
    assert result[3] == approx(7.40000e-01, 0.2)
    assert result[4] == approx(8.41773e-01, 0.2)
    assert result[5] == 0
    assert result[6] == 0
    assert result[7] == 0
    assert is_jsonable(dict(zip(av.feature_labels(), result)))
    assert result.ndim == 1


def test_raytracing_histogram(hkust_structure):
    """Ensure that the featurizer works as expected."""
    rth = RayTracingHistogram()
    assert len(rth.feature_labels()) == 1000
    assert len(rth.citations()) == 2
    features = rth.featurize(hkust_structure)
    assert len(features) == 1000
    assert features[0] >= 1.0
    assert is_jsonable(dict(zip(rth.feature_labels(), features)))
    assert features.ndim == 1


def test_psd(hkust_structure):
    """Ensure that the featurizer works as expected."""
    psd = PoreSizeDistribution()
    assert len(psd.feature_labels()) == 1000
    assert len(psd.citations()) == 2
    features = psd.featurize(hkust_structure)
    assert len(features) == 1000
    assert features[0] == 0
    assert is_jsonable(dict(zip(psd.feature_labels(), features)))
    assert features.ndim == 1



# -*- coding: utf-8 -*-
"""Partial charge statistics featurizer."""
from typing import List, Tuple, Union

import numpy as np
from pymatgen.core import IStructure, Structure

from mofdscribe.featurizers.base import MOFBaseFeaturizer
from mofdscribe.featurizers.utils.aggregators import ARRAY_AGGREGATORS
from mofdscribe.featurizers.utils.eqeq import get_eqeq_charges
from mofdscribe.featurizers.utils.extend import operates_on_istructure, operates_on_structure

__all__ = ["PartialChargeStats"]


@operates_on_istructure
@operates_on_structure
class PartialChargeStats(MOFBaseFeaturizer):
    """Compute partial charges using the EqEq charge equilibration method [Ongari2019]_.

    Then derive a fix-length feature vector from the partial charges using
    aggregative statistics.

    They have, for instance, been used as "maximum positive charge" and "mininum
    negative charge" in `Moosavi et al. (2020)
    <https://www.nature.com/articles/s41467-020-17755-8>`_
    """

    def __init__(
        self, aggregations: Tuple[str] = ("max", "min", "std"), primitive: bool = True
    ) -> None:
        """
        Initialize the PartialChargeStats featurizer.

        Args:
            aggregations (Tuple[str]): Aggregations to compute.
                For available methods, see :py:obj:`mofdscribe.featurizers.utils.aggregators.ARRAY_AGGREGATORS`.
                Defaults to ("max", "min", "std").
            primitive (bool): If True, the structure is reduced to its primitive
                form before the descriptor is computed. Defaults to True.
        """
        self.aggregations = aggregations
        super().__init__(primitive=primitive)

    def feature_labels(self) -> List[str]:
        return [f"chargestat_{agg}" for agg in self.aggregations]

    def _featurize(self, s: Union[Structure, IStructure]) -> np.ndarray:
        if isinstance(s, Structure):
            s = IStructure.from_sites(s.sites)
        _, results = get_eqeq_charges(s)
        aggregated = [ARRAY_AGGREGATORS[agg](results) for agg in self.aggregations]

        return np.array(aggregated)

    def citations(self) -> List[str]:
        return [
            "@article{Ongari2018,"
            "doi = {10.1021/acs.jctc.8b00669},"
            "url = {https://doi.org/10.1021/acs.jctc.8b00669},"
            "year = {2018},"
            "month = nov,"
            "publisher = {American Chemical Society ({ACS})},"
            "volume = {15},"
            "number = {1},"
            "pages = {382--401},"
            "author = {Daniele Ongari and Peter G. Boyd and Ozge Kadioglu "
            "and Amber K. Mace and Seda Keskin and Berend Smit},"
            "title = {Evaluating Charge Equilibration Methods "
            "To Generate Electrostatic Fields in Nanoporous Materials},"
            "journal = {Journal of Chemical Theory and Computation}"
            "}",
            "@article{Wilmer2012,"
            "doi = {10.1021/jz3008485},"
            "url = {https://doi.org/10.1021/jz3008485},"
            "year = {2012},"
            "month = aug,"
            "publisher = {American Chemical Society ({ACS})},"
            "volume = {3},"
            "number = {17},"
            "pages = {2506--2511},"
            "author = {Christopher E. Wilmer and Ki Chul Kim and Randall Q. Snurr},"
            "title = {An Extended Charge Equilibration Method},"
            "journal = {The Journal of Physical Chemistry Letters}"
            "}",
        ]

    def implementors(self):
        return ["Kevin Maik Jablonka", "Daniele Ongari", "Christopher Wilmer"]



# -*- coding: utf-8 -*-
"""Test the RACs featurizer."""
from collections import OrderedDict

import numpy as np
from pymatgen.core import IStructure

from mofdscribe.featurizers.chemistry._fragment import get_bb_indices
from mofdscribe.featurizers.chemistry.racs import RACS, _get_racs_for_bbs
from mofdscribe.featurizers.utils.structure_graph import get_structure_graph

from ..helpers import is_jsonable


def test_racs(hkust_structure, irmof_structure):
    """Make sure that the featurization works for typical MOFs and the number of features is as expected."""
    for structure in [hkust_structure, irmof_structure]:
        featurizer = RACS()
        feats = featurizer.featurize(structure)
        sg = get_structure_graph(IStructure.from_sites(structure), featurizer.bond_heuristic)
        racs = {}
        bb_indices = get_bb_indices(sg)
        for bb in featurizer._bbs:
            v = _get_racs_for_bbs(
                bb_indices[bb],
                sg,
                featurizer.attributes,
                featurizer.scopes,
                featurizer.prop_agg,
                featurizer.corr_agg,
                featurizer.bb_agg,
                bb,
            )
            racs.update(v)
            # There are no functional groups in those MOFs, so we except nans for this "scope"
            if ("functional" not in bb) and ("linker" in bb):
                assert np.isnan(np.array(list(v.values()))).sum() == 0
            elif "functional" in bb:
                assert np.isnan(np.array(list(v.values()))).sum() == len(v)
        racs_ordered = OrderedDict(sorted(racs.items()))

        assert list(racs_ordered.keys()) == featurizer.feature_labels()

    # assert len(featurizer.feature_labels()) == 120
    assert len(featurizer.citations()) == 2
    assert is_jsonable(dict(zip(featurizer.feature_labels(), feats)))
    assert feats.ndim == 1


def test_racs_functional(irmof_structure, abacuf_structure, floating_structure):
    # ABACUF doesn't have linkers with a core. It is simply HCOO
    for structure in [abacuf_structure]:
        featurizer = RACS(primitive=False)  # because also the linker structure is not primitive
        feats = featurizer.featurize(structure)
        # assert len(feats) == 4 * 3 * 8 * 5  # 4 properties, 3 scopes, 8 aggregations, 5 bb types
        sg = get_structure_graph(IStructure.from_sites(structure), featurizer.bond_heuristic)
        racs = {}
        bb_indices = get_bb_indices(sg)
        for bb in featurizer._bbs:
            v = _get_racs_for_bbs(
                bb_indices[bb],
                sg,
                featurizer.attributes,
                featurizer.scopes,
                featurizer.prop_agg,
                featurizer.corr_agg,
                featurizer.bb_agg,
                bb,
            )
            racs.update(v)
            # we classify the "O" as a functional group
            assert np.isnan(np.array(list(v.values()))).sum() == 0
        racs_ordered = OrderedDict(sorted(racs.items()))
        assert list(racs_ordered.keys()) == featurizer.feature_labels()

    # assert len(featurizer.feature_labels()) == 120
    assert len(featurizer.citations()) == 2
    assert is_jsonable(dict(zip(featurizer.feature_labels(), feats)))
    assert feats.ndim == 1

    floating_feats = featurizer.featurize(floating_structure)
    irmof_feats = featurizer.featurize(irmof_structure)

    assert np.allclose(floating_feats, irmof_feats, rtol=0.05, equal_nan=True)



# -*- coding: utf-8 -*-
"""Test the helper functions for dealing with substructures."""
from pymatgen.core import Structure

from mofdscribe.featurizers.utils.substructures import get_metal_indices, select_elements


def test_get_metal_indices(irmof_structure):
    """Ensure that we get the correct indices for the metal atoms."""
    s = irmof_structure
    metal_indices = get_metal_indices(s)
    assert len(metal_indices) == 32
    assert metal_indices[0] == 0
    assert metal_indices[-1] == 31


def test_select_elements(irmof_structure):
    """Ensure that we can extract a correct substructure."""
    s = irmof_structure
    s = select_elements(s, "Zn")
    assert isinstance(s, Structure)
    assert len(s) == 32
    assert s[0].specie.symbol == "Zn"



# -*- coding: utf-8 -*-
"""Compute local structure order parameters for a fragment."""
from typing import List, Optional, Tuple, Union

import numpy as np
from matminer.featurizers.base import BaseFeaturizer
from pymatgen.analysis.local_env import LocalStructOrderParams
from pymatgen.core import IMolecule, IStructure, Molecule, Structure
from pymatgen.core.periodic_table import DummySpecies

from mofdscribe.featurizers.utils.extend import (
    operates_on_imolecule,
    operates_on_istructure,
    operates_on_molecule,
    operates_on_structure,
)

_default_types = (
    "cn",
    "tet",
    "oct",
    "bcc",
    "sq_pyr",
    "sq_pyr_legacy",
    "tri_bipyr",
    "sq_bipyr",
    "oct_legacy",
    "tri_plan",
    "sq_plan",
    "pent_plan",
    "tri_pyr",
    "pent_pyr",
    "hex_pyr",
    "pent_bipyr",
    "hex_bipyr",
    "T",
    "cuboct",
    "oct_max",
    "tet_max",
    "tri_plan_max",
    "sq_plan_max",
    "pent_plan_max",
    "cuboct_max",
    "bent",
    "see_saw_rect",
    "hex_plan_max",
    "sq_face_cap_trig_pris",
)


@operates_on_structure
@operates_on_istructure
@operates_on_molecule
@operates_on_imolecule
class LSOP(BaseFeaturizer):
    """Compute shape parameters for a fragment.

    The fragments can be a molecule or a molecule that only contains important part (e.g. binding sites)
    of a molecule. The shape parameters are then supposed to quantify the shape of the fragment.
    For instance, a triangular molecule will have a `tri_plan` parameter close to 1.

    While there is a site-based LSOP featurizer in matminer there is none that uses LSOP
    to quantify the shape of some fragment. This featurizers just does that.

    It does so by placing a dummy site at the center of mass
    of the fragment and then computes the LSOP considering all other sites as neighbors.
    """

    def __init__(
        self, types: Tuple[str] = _default_types, parameters: Optional[List[dict]] = None
    ) -> None:
        """Initialize the featurizer.

        Args:
            types (Tuple[str]): The types of LSOP to compute.
                For the full list of types
                see: :py:attr:`pymatgen.analysis.local_env.LocalStructOrderParams. __supported_types`.
                Defaults to: ["tet", "oct", "bcc", "sq_pyr", "sq_pyr_legacy", "tri_bipyr", "sq_bipyr",
                "oct_legacy", "tri_plan", "sq_plan", "pent_plan", "tri_pyr", "pent_pyr", "hex_pyr",
                "pent_bipyr", "hex_bipyr", "T", "cuboct", "oct_max", "tet_max", "tri_plan_max",
                "sq_plan_max", "pent_plan_max", "cuboct_max", "bent", "see_saw_rect", "hex_plan_max",
                "sq_face_cap_trig_pris"]
            parameters(List[dict], optional): The parameters to pass to the LocalStructOrderParams object.
        """
        self._lsop = LocalStructOrderParams(types, parameters)
        self.types = types

    def feature_labels(self) -> List[str]:
        return [f"lsop_{val}" for val in self.types]

    def featurize(self, s: Union[Structure, IStructure, Molecule, IMolecule]) -> np.ndarray:
        molecule = Molecule.from_sites(s.sites)
        com = molecule.center_of_mass
        orginal_len = len(molecule)
        molecule.append(DummySpecies(), com)
        lsop = self._lsop.get_order_parameters(
            molecule, len(molecule) - 1, indices_neighs=np.arange(orginal_len)
        )
        lsop = [f if f is not None else np.nan for f in lsop]
        return np.array(lsop)

    def implementors(self) -> List[str]:
        return ["Kevin Maik Jablonka"]

    def citations(self) -> List[str]:
        return [
            "@article{Zimmermann2020,"
            "doi = {10.1039/c9ra07755c},"
            "url = {https://doi.org/10.1039/c9ra07755c},"
            "year = {2020},"
            "publisher = {Royal Society of Chemistry ({RSC})},"
            "volume = {10},"
            "number = {10},"
            "pages = {6063--6081},"
            "author = {Nils E. R. Zimmermann and Anubhav Jain},"
            "title = {Local structure order parameters and site fingerprints"
            " for quantification of coordination environment and crystal structure similarity},"
            "journal = {{RSC} Advances}"
            "}"
        ]



# -*- coding: utf-8 -*-
"""Compute statistics of persistent images for MOFs."""
from typing import List, Optional, Tuple, Union

import numpy as np
from pymatgen.core import IMolecule, IStructure, Molecule, Structure

from mofdscribe.featurizers.base import MOFBaseFeaturizer
from mofdscribe.featurizers.utils import flatten
from mofdscribe.featurizers.utils.extend import (
    operates_on_imolecule,
    operates_on_istructure,
    operates_on_molecule,
    operates_on_structure,
)

from ._tda_helpers import get_diagrams_for_structure, persistent_diagram_stats


@operates_on_imolecule
@operates_on_molecule
@operates_on_istructure
@operates_on_structure
class PHStats(MOFBaseFeaturizer):
    """Featurizer that computes statistics of persistent images.

    Compute a fixed-length vector of topological descriptors for a structure by
    summarizing the persistence diagrams of the structure (or substructure)
    using aggegrations such as `min`, `max`, `mean`, and `std`.

    The descriptors can be computed over the full structure or substructures of
    certain atom types.
    """

    def __init__(
        self,
        atom_types: Tuple[str] = (
            "C-H-N-O",
            "F-Cl-Br-I",
            "Cu-Mn-Ni-Mo-Fe-Pt-Zn-Ca-Er-Au-Cd-Co-Gd-Na-Sm-Eu-Tb-V"
            "-Ag-Nd-U-Ba-Ce-K-Ga-Cr-Al-Li-Sc-Ru-In-Mg-Zr-Dy-W-Yb-Y-"
            "Ho-Re-Be-Rb-La-Sn-Cs-Pb-Pr-Bi-Tm-Sr-Ti-Hf-Ir-Nb-Pd-Hg-"
            "Th-Np-Lu-Rh-Pu",
        ),
        compute_for_all_elements: bool = True,
        dimensions: Tuple[int] = (1, 2),
        min_size: int = 20,
        aggregation_functions: Tuple[str] = ("min", "max", "mean", "std"),
        periodic: bool = False,
        no_supercell: bool = False,
        primitive: bool = False,
        alpha_weight: Optional[str] = None,
    ) -> None:
        """Initialize the PHStats object.

        Args:
            atom_types (tuple): Atoms that are used to create substructures
                for which the persistent homology statistics are computed.
                Defaults to ( "C-H-N-O", "F-Cl-Br-I",
                "Cu-Mn-Ni-Mo-Fe-Pt-Zn-Ca-Er-Au-Cd-Co-Gd-Na-Sm-Eu-Tb-V-Ag-Nd-U-Ba-Ce-K-Ga-
                Cr-Al-Li-Sc-Ru-In-Mg-Zr-Dy-W-Yb-Y-Ho-Re-Be-Rb-La-Sn-Cs-Pb-Pr-Bi-Tm-Sr-Ti-
                Hf-Ir-Nb-Pd-Hg-Th-Np-Lu-Rh-Pu", ).
            compute_for_all_elements (bool): Compute descriptor for original structure with all atoms.
                Defaults to True.
            dimensions (Tuple[int]): Dimensions of topological features to consider.
                Defaults to (1, 2).
            min_size (int): Minimum supercell size (in Angstrom).
                Defaults to 20.
            aggregation_functions (Tuple[str]): Methods used to combine the properties.
                See `mofdscribe.featurizers.utils.aggregators.ARRAY_AGGREGATORS` for available options.
                Defaults to ("min", "max", "mean", "std").
            periodic (bool): If true, then periodic Euclidean is used in the analysis (experimental!).
                Defaults to False.
            no_supercell (bool): If true, then the supercell is not created.
                Defaults to False.
            primitive (bool): If True, the structure is reduced to its primitive
                form before the descriptor is computed. Defaults to False.
            alpha_weight (Optional[str]): If specified, the use weighted alpha shapes,
                i.e., replacing the points with balls of varying radii.
                For instance `atomic_radius_calculated` or `van_der_waals_radius`.
        """
        atom_types = [] if atom_types is None else atom_types
        self.elements = atom_types
        self.atom_types = (
            list(atom_types) + ["all"] if compute_for_all_elements else list(atom_types)
        )
        self.compute_for_all_elements = compute_for_all_elements
        self.dimensions = dimensions
        self.min_size = min_size
        self.aggregation_functions = aggregation_functions
        self.periodic = periodic
        self.no_supercell = no_supercell
        self.alpha_weight = alpha_weight
        super().__init__(primitive=primitive)

    def _get_feature_labels(self) -> List[str]:
        labels = []
        for atom_type in self.atom_types:
            for dim in self.dimensions:
                for parameter in ("birth", "death", "persistence"):
                    for aggregation in self.aggregation_functions:
                        labels.append(f"phstats_{atom_type}_dim{dim}_{parameter}_{aggregation}")

        return labels

    def feature_labels(self) -> List[str]:
        return self._get_feature_labels()

    def _featurize(
        self, structure: Union[Structure, IStructure, Molecule, IMolecule]
    ) -> np.ndarray:
        res = get_diagrams_for_structure(
            structure,
            self.elements,
            self.compute_for_all_elements,
            self.min_size,
            periodic=self.periodic,
            no_supercell=self.no_supercell,
            alpha_weighting=self.alpha_weight,
        )

        flat_results = []
        for atom_type in self.atom_types:
            for dim in self.dimensions:

                dimname = f"dim{dim}"
                stats = persistent_diagram_stats(
                    res[atom_type][dimname], self.aggregation_functions
                )
                flat_stats = list(flatten(stats).values())
                flat_results.extend(flat_stats)

        return np.array(flat_results)

    def citations(self):
        return [
            "@article{doi:10.1021/acs.jpcc.0c01167,"
            "author = {Krishnapriyan, Aditi S. and Haranczyk, Maciej and Morozov, Dmitriy},"
            "title = {Topological Descriptors Help Predict Guest Adsorption in Nanoporous Materials},"
            "journal = {The Journal of Physical Chemistry C},"
            "volume = {124},"
            "number = {17},"
            "pages = {9360-9368},"
            "year = {2020},"
            "doi = {10.1021/acs.jpcc.0c01167},"
            "}",
        ]

    def implementors(self):
        return ["Kevin Maik Jablonka", "Aditi Krishnapriyan"]



# -*- coding: utf-8 -*-
"""Test that the fragment LSOP featurizer works."""

from mofdscribe.featurizers.bu.lsop_featurizer import LSOP


def test_lsop_featurizer(triangle_structure):
    """Test that the LSOP featurizer works."""
    featurizer = LSOP(types=["tri_plan", "sq_pyr", "cn"])
    features = featurizer.featurize(triangle_structure)
    assert features.shape == (3,)
    assert len(features) == len(featurizer.feature_labels())
    assert features[0] > 0.5
    assert features[1] < 0.1
    assert features[2] == 3



# -*- coding: utf-8 -*-
"""Classes that help performing cross-validation.

Our splitters attempt to reduce any potential for data leakage by using grouping by default--
and prioritizing grouping over stratficiation or exactly matching the requested train test ratio.

See also the `sklearn docs
<https://scikit-learn.org/stable/modules/cross_validation.html#cross-validation-iterators-for-grouped-data>`_.

.. warning::

    Due to the grouping operations, the train/test ratios the methods produce will not exactly
    match the one you requested.
    For this reason, please get the length of the train/test/valid indices the methods produce.
"""
from typing import Any, Callable, Collection, Dict, List, Optional, Tuple, Union

import numpy as np
from loguru import logger
from sklearn.model_selection import GroupKFold, KFold, StratifiedGroupKFold, StratifiedKFold

from .utils import (
    check_fraction,
    downsample_splits,
    grouped_stratified_train_test_partition,
    grouped_train_valid_test_partition,
    is_categorical,
    kennard_stone_sampling,
    no_group_warn,
    pca_kmeans,
    quantile_binning,
    sort_arrays_by_len,
    stratified_train_test_partition,
)
from ..datasets.dataset import AbstractStructureDataset

__all__ = (
    "DensitySplitter",
    "HashSplitter",
    "TimeSplitter",
    "BaseSplitter",
    "KennardStoneSplitter",
    "ClusterSplitter",
    "LOCOCV",
    "ClusterStratifiedSplitter",
)


class BaseSplitter:
    """A :code:`BaseSplitter` implements the basic logic for dataset partition as well as k-fold cross-validation.

    Methods that inherit from this class typically implement the

        * :code: `_get_stratification_col`: Should return an ArrayLike object of floats, categories, or ints.
            If it is categorical data, the :code:`BaseSplitter` will handle the discretization.
        * :code: `_get_groups`: Should return an ArrayLike object of categories (integers or strings)

    methods.
    Internally, the :code:`BaseSplitter` uses those to group and/or stratify the splits.
    """

    # Those variables are needed to automatically set the number of groups
    # if the users does not set them (default behavior).
    _grouping_q = None
    _set_grouping = False

    def __init__(
        self,
        ds: AbstractStructureDataset,
        shuffle: bool = True,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        sample_frac: Optional[float] = 1.0,
        stratification_col: Optional[Union[str, np.typing.ArrayLike]] = None,
        center: callable = np.median,
        q: Collection[float] = (0, 0.25, 0.5, 0.75, 1),
        sort_by_len: bool = True,
    ):
        """Initialize a BaseSplitter.

        Args:
            ds (AbstractStructureDataset): A structure dataset.
                The :code:`BaseSplitter` only requires the length magic method to be implemented.
                However, other splitters might require additional methods.
            shuffle (bool): If True, perform a shuffled split.
                Defaults to True.
            random_state (Optional[Union[int, np.random.RandomState]], optional):
                Random state for the shuffling. Defaults to None.
            sample_frac (Optional[float], optional):
                This can be used for downsampling. It will randomly select a subset of
                indices from all indices *before* splittings. For instance :code:`sample_frac=0.8`
                will randomly select 80% of the indices before splitting.
                Defaults to 1.0.
            stratification_col (Union[str, np.typing.ArrayLike], optional): Data used for stratification.
                If it is categorical (see :py:meth:`mofdscribe.splitters.utils.is_categorical`)
                then we directly use it for stratification. Otherwise, we use quantile binning.
                Defaults to None.
            center (callable): Aggregation function to compute a measure of centrality
                of all the points in a group such that this can then be used for stratification.
                This is only used for continuos inputs. For categorical inputs, we always use
                the mode. Defaults to np.median.
            q (Collection[float], optional): List of quantiles used for quantile binning.
                Defaults to (0, 0.25, 0.5, 0.75, 1).
            sort_by_len (bool): If True, sort the splits by length.
                (Applies to the train/test/valid and train/test splits). Defaults to True.
        """
        self._ds = ds
        self._shuffle = shuffle
        self._random_state = random_state
        self._len = len(ds)
        self._sample_frac = sample_frac
        self._stratification_col = stratification_col
        self._center = center
        self._q = q
        self._sort_by_len = sort_by_len

        logger.debug(
            f"Splitter settings | shuffle {self._shuffle}, "
            f"random state {self._random_state}, sample frac {self._sample_frac}, q {self._q}"
        )

    def _get_idxs(self):
        """Return an array of indices. Length equals to the length of the dataset."""
        idx = np.arange(self._len)
        return idx

    def train_test_split(self, frac_train: float = 0.7) -> Tuple[Collection[int], Collection[int]]:
        """Perform a train/test partition.

        Args:
            frac_train (float): Fraction of the data to use for the training set.
                Defaults to 0.7.

        Returns:
            Tuple[Collection[int], Collection[int]]: Train indices, test indices
        """
        if self._grouping_q is None or self._set_grouping:
            self._set_grouping = True
            self._grouping_q = np.linspace(0, 1, 3)
        check_fraction(train_fraction=frac_train, valid_fraction=0, test_fraction=1 - frac_train)
        groups = self._get_groups()
        stratification_col = self._get_stratification_col()
        idx = self._get_idxs()
        no_group_warn(groups)
        if groups is not None:
            if stratification_col is not None:
                logger.debug("Using grouped stratified partition")
                train_idx, _, test_index = grouped_stratified_train_test_partition(
                    stratification_col[idx],
                    groups[idx],
                    frac_train,
                    0,
                    1 - frac_train,
                    shuffle=self._shuffle,
                    random_state=self._random_state,
                    center=self._center,
                    q=self._q,
                )
            else:
                logger.debug("Using grouped partition")
                train_idx, _, test_index = grouped_train_valid_test_partition(
                    groups[idx],
                    frac_train,
                    0,
                    1 - frac_train,
                    shuffle=self._shuffle,
                    random_state=self._random_state,
                )

        else:
            stratification_col = stratification_col[idx] if stratification_col is not None else None
            logger.debug("Using stratified partition")
            train_idx, _, test_index = stratified_train_test_partition(
                self._get_idxs(),
                stratification_col,
                train_size=frac_train,
                valid_size=0,
                test_size=1 - frac_train,
                shuffle=self._shuffle,
                random_state=self._random_state,
                q=self._q,
            )

        if self._sample_frac < 1:
            return sort_arrays_by_len(
                downsample_splits([train_idx, test_index], self._sample_frac), self._sort_by_len
            )
        return sort_arrays_by_len([train_idx, test_index], self._sort_by_len)

    def train_valid_test_split(
        self, frac_train: float = 0.7, frac_valid: float = 0.1
    ) -> Tuple[Collection[int], Collection[int], Collection[int]]:
        """Perform a train/valid/test partition.

        Args:
            frac_train (float): Fraction of data to use for the training set.
                Defaults to 0.7.
            frac_valid (float): Fraction of data to use for the validation set.
                Defaults to 0.1.

        Returns:
            Tuple[Collection[int], Collection[int], Collection[int]]: Training, validation, test set.
        """
        if self._grouping_q is None or self._set_grouping:
            self._set_grouping = True
            self._grouping_q = np.linspace(0, 1, 4)
        check_fraction(
            train_fraction=frac_train,
            valid_fraction=frac_valid,
            test_fraction=1 - frac_train - frac_valid,
        )
        groups = self._get_groups()
        stratification_col = self._get_stratification_col()
        idx = self._get_idxs()
        no_group_warn(groups)

        if groups is not None:
            if stratification_col is not None:
                logger.debug("Using grouped stratified partition")
                train_idx, valid_idx, test_index = grouped_stratified_train_test_partition(
                    stratification_col[idx],
                    groups[idx],
                    frac_train,
                    frac_valid,
                    1 - frac_train - frac_valid,
                    shuffle=self._shuffle,
                    random_state=self._random_state,
                    center=self._center,
                    q=self._q,
                )
            else:
                logger.debug("Using grouped  partition")
                train_idx, valid_idx, test_index = grouped_train_valid_test_partition(
                    groups[idx],
                    frac_train,
                    frac_valid,
                    1 - frac_train - frac_valid,
                    shuffle=self._shuffle,
                    random_state=self._random_state,
                )
        else:
            logger.debug("Using stratified partition")
            stratification_col = stratification_col[idx] if stratification_col is not None else None
            train_idx, valid_idx, test_index = stratified_train_test_partition(
                self._get_idxs(),
                stratification_col,
                train_size=frac_train,
                valid_size=frac_valid,
                test_size=1 - frac_train - frac_valid,
                shuffle=self._shuffle,
                random_state=self._random_state,
                q=self._q,
            )
        if self._sample_frac < 1:
            return sort_arrays_by_len(
                downsample_splits([train_idx, valid_idx, test_index], self._sample_frac),
                self._sort_by_len,
            )

        return sort_arrays_by_len([train_idx, valid_idx, test_index], self._sort_by_len)

    def k_fold(self, k: int = 5) -> Tuple[Collection[int], Collection[int]]:
        """Peform k-fold crossvalidation.

        Args:
            k (int): Number of folds. Defaults to 5.

        Yields:
            Tuple[Collection[int], Collection[int]]: Train indices, test indices.
        """
        if self._grouping_q is None or self._set_grouping:
            self._set_grouping = True
            self._grouping_q = np.linspace(0, 1, k + 1)
        groups = self._get_groups()
        stratification_col = self._get_stratification_col()
        no_group_warn(groups)
        idx = self._get_idxs()

        groups = groups[idx] if groups is not None else None
        stratification_col = stratification_col[idx] if stratification_col is not None else None

        if stratification_col is not None:
            if not is_categorical(stratification_col):
                stratification_col = quantile_binning(stratification_col, self._q)

            if groups is not None:
                kfold = StratifiedGroupKFold(
                    n_splits=k, shuffle=self._shuffle, random_state=self._random_state
                )
            else:
                kfold = StratifiedKFold(
                    n_splits=k, shuffle=self._shuffle, random_state=self._random_state
                )

        else:
            # this is not shuffled?
            if groups is not None:
                kfold = GroupKFold(n_splits=k)
            else:
                kfold = KFold(n_splits=k)

        if groups is not None:
            for train_index, test_index in kfold.split(idx, y=stratification_col, groups=groups):
                if self._shuffle:
                    np.random.shuffle(train_index)
                    np.random.shuffle(test_index)
                if self._sample_frac < 1:
                    yield downsample_splits([train_index, test_index], self._sample_frac)
                else:
                    yield train_index, test_index
        else:
            for train_index, test_index in kfold.split(idx, y=stratification_col):
                if self._shuffle:
                    np.random.shuffle(train_index)
                    np.random.shuffle(test_index)
                if self._sample_frac < 1:
                    yield downsample_splits([train_index, test_index], self._sample_frac)
                else:
                    yield train_index, test_index

    def _get_groups(self) -> Collection[Union[int, str]]:
        return None

    def _get_stratification_col(self) -> Collection[Union[int, float]]:
        if isinstance(self._stratification_col, str):
            return self._ds._df[self._stratification_col].values
        else:
            return self._stratification_col


class HashSplitter(BaseSplitter):
    """Splitter that uses Weisfeiller-Lehman graph hashes [WL]_ to split the data in more stringent ways.

    Note that the hashes we use do not allow for a meaningful measure of
    similarity. That is, there is no way to measure the distance between two strings.
    The only meaningful measure is if they are identical or not.


    .. note::

        Weisfeiller-Lehman graph hashes do not give a guarantee for graph-isomorphism.
        That is, there might be identical hashes that do not correspond to isomorphic graphs.

    .. note::

        There are certain graphs that a Weisfeiller-Lehman test cannot distinguish [Bouritsas]_.

    .. note::

        We speak about Weisfeiller-Lehman hashes as they are the defaults for the mofdscribe datasets.
        However, you can also overwrite this method with a custom hashing function.
    """

    def __init__(
        self,
        ds: AbstractStructureDataset,
        hash_type: str = "undecorated_scaffold_hash",
        shuffle: bool = True,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        sample_frac: Optional[float] = 1.0,
        stratification_col: Optional[Union[str, np.typing.ArrayLike]] = None,
        center=np.median,
        q: Collection[float] = (0, 0.25, 0.5, 0.75, 1),
        sort_by_len: bool = True,
    ) -> None:
        """Initialize a HashSplitter.

        Args:
            ds (AbstractStructureDataset): A structure dataset.
                The :code:`BaseSplitter` only requires the length magic method to be implemented.
                However, other splitters might require additional methods.
            hash_type (str): Hash type to use. Must be one of the
                following:
                * undecorated_scaffold_hash
                * decorated_graph_hash
                * decorated_scaffold_hash
                * undecorated_graph_hash
                Defaults to "undecorated_scaffold_hash".
            shuffle (bool): If True, perform a shuffled split.
                Defaults to True.
            random_state (Union[int, np.random.RandomState], optional):
                Random state for the shuffling. Defaults to None.
            sample_frac (float, optional):
                This can be used for downsampling. It will randomly select a subset of
                indices from all indices *before* splittings. For instance :code:`sample_frac=0.8`
                will randomly select 80% of the indices before splitting.
                Defaults to 1.0.
            stratification_col (Union[str, np.typing.ArrayLike], optional): Data used for stratification.
                If it is categorical (see :py:meth:`mofdscribe.splitters.utils.is_categorical`)
                then we directly use it for stratification. Otherwise, we use quantile binning.
                Defaults to None.
            center (callable, optional): Aggregation function to compute a measure of centrality
                of all the points in a group such that this can then be used for stratification.
                This is only used for continuos inputs. For categorical inputs, we always use
                the mode. Defaults to np.median.
            q (Collection[float], optional): List of quantiles used for quantile binning.
                Defaults to (0, 0.25, 0.5, 0.75, 1).
            sort_by_len (bool): If True, sort the splits by length.
                (Applies to the train/test/valid and train/test splits). Defaults to True.
        """
        self.hash_type = hash_type
        super().__init__(
            ds,
            shuffle=shuffle,
            random_state=random_state,
            sample_frac=sample_frac,
            stratification_col=stratification_col,
            center=center,
            q=q,
            sort_by_len=sort_by_len,
        )

    def _get_hashes(self) -> Collection[str]:
        """Retrieve the list of hashes from the dataset

        Raises:
            ValueError: If the hash type is not one of the following:
                * undecorated_scaffold_hash
                * decorated_graph_hash
                * decorated_scaffold_hash
                * undecorated_graph_hash

        Returns:
            Collection[str]: list of hashes
        """
        number_of_points = len(self._ds)
        if self.hash_type == "undecorated_scaffold_hash":
            hashes = self._ds.get_undecorated_scaffold_hashes(range(number_of_points))
        elif self.hash_type == "decorated_graph_hash":
            hashes = self._ds.get_decorated_graph_hashes(range(number_of_points))
        elif self.hash_type == "decorated_scaffold_hash":
            hashes = self._ds.get_decorated_scaffold_hashes(range(number_of_points))
        elif self.hash_type == "undecorated_graph_hash":
            hashes = self._ds.get_undecorated_graph_hashes(range(number_of_points))
        else:
            raise ValueError(f"Unknown hash type: {self.hash_type}")

        return hashes

    def _get_groups(self) -> Collection[int]:
        return self._get_hashes()


class DensitySplitter(BaseSplitter):
    """Splitter that uses the density of the structures to split the data.

    For this, we sort structures according to their density and then group the based on the density.
    You can modify the number of groups using the :attr:`density_q` parameter, those values indicate
    the quantiles which we use for the grouping.

    This ensures that the validation is quite stringent as the different folds will have different densities.

    The motivations for doing this are:

        * density is often one of the most important descriptors for gas uptake properties.

        * there is often is a very large difference in density distribution
            between hypothetical and experimental databases.
    """

    def __init__(
        self,
        ds: AbstractStructureDataset,
        density_q: Optional[Collection[float]] = None,
        shuffle: bool = True,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        sample_frac: Optional[float] = 1.0,
        stratification_col: Optional[Union[str, np.typing.ArrayLike]] = None,
        center: callable = np.median,
        q: Collection[float] = (0, 0.25, 0.5, 0.75, 1),
        sort_by_len: bool = True,
    ) -> None:
        """Initialize the DensitySplitter class.

        Args:
            ds (AbstractStructureDataset): A structure dataset.
                The :code:`BaseSplitter` only requires the length magic method to be implemented.
                However, other splitters might require additional methods.
            density_q (Collection[float], optional): List of quantiles used for quantile binning for the density.
                Defaults to None. If None, then we use two bins for test/train split, three for
                validation/train/test split and k for k-fold.
            shuffle (bool): If True, perform a shuffled split.
                Defaults to True.
            random_state (Union[int, np.random.RandomState], optional):
                Random state for the shuffling. Defaults to None.
            sample_frac (float, optional):
                This can be used for downsampling. It will randomly select a subset of
                indices from all indices *before* splittings. For instance :code:`sample_frac=0.8`
                will randomly select 80% of the indices before splitting.
                Defaults to 1.0.
            stratification_col (Union[str, np.typing.ArrayLike], optional): Data used for stratification.
                If it is categorical (see :py:meth:`mofdscribe.splitters.utils.is_categorical`)
                then we directly use it for stratification. Otherwise, we use quantile binning.
                Defaults to None.
            center (callable): Aggregation function to compute a measure of centrality
                of all the points in a group such that this can then be used for stratification.
                This is only used for continuos inputs. For categorical inputs, we always use
                the mode. Defaults to np.median.
            q (Collection[float], optional): List of quantiles used for quantile binning.
                Defaults to (0, 0.25, 0.5, 0.75, 1]. Defaults to [0, 0.25, 0.5, 0.75, 1).
            sort_by_len (bool): If True, sort the splits by length.
                (Applies to the train/test/valid and train/test splits). Defaults to True.
        """
        self._grouping_q = density_q
        super().__init__(
            ds=ds,
            shuffle=shuffle,
            random_state=random_state,
            sample_frac=sample_frac,
            stratification_col=stratification_col,
            center=center,
            q=q,
            sort_by_len=sort_by_len,
        )

    def _get_groups(self) -> Collection[int]:
        return quantile_binning(self._ds.get_densities(range(len(self._ds))), self._grouping_q)


class TimeSplitter(BaseSplitter):
    """This splitter sorts structures according to their publication date.

    That is, the training set will contain structures that are "older" (have
    been discovered earlier) than the ones in the test set.
    This can mimick real-life model development conditions [MoleculeNet]_.

    It has for instance also be used with ICSD data in [Palizhati]_
    and been the focus of [Sheridan]_.

    .. seealso:

        * The `mp-time-split <https://github.com/sparks-baird/mp-time-split>`_ package
        provides similar functionality for data from the materials project.
    """

    def __init__(
        self,
        ds: AbstractStructureDataset,
        year_q: Optional[Collection[float]] = None,
        shuffle: bool = True,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        sample_frac: Optional[float] = 1.0,
        stratification_col: Optional[Union[str, np.typing.ArrayLike]] = None,
        center: callable = np.median,
        q: Collection[float] = (0, 0.25, 0.5, 0.75, 1),
        sort_by_len: bool = True,
    ) -> None:
        """Initialize the TimeSplitter class.

        Args:
            ds (AbstractStructureDataset): A structure dataset.
                The :code:`BaseSplitter` only requires the length magic method to be implemented.
                However, other splitters might require additional methods.
            year_q (Collection[float]): List of quantiles used for quantile binning on the years.
                Defaults to None. If None, then we use two bins for test/train split, three for
                validation/train/test split and k for k-fold.
            shuffle (bool): If True, perform a shuffled split.
                Defaults to True.
            random_state (Union[int, np.random.RandomState], optional):
                Random state for the shuffling. Defaults to None.
            sample_frac (float, optional):
                This can be used for downsampling. It will randomly select a subset of
                indices from all indices *before* splittings. For instance :code:`sample_frac=0.8`
                will randomly select 80% of the indices before splitting.
                Defaults to 1.0.
            stratification_col (Union[str, np.typing.ArrayLike], optional): Data used for stratification.
                If it is categorical (see :py:meth:`mofdscribe.splitters.utils.is_categorical`)
                then we directly use it for stratification. Otherwise, we use quantile binning.
                Defaults to None.
            center (callable): Aggregation function to compute a measure of centrality
                of all the points in a group such that this can then be used for stratification.
                This is only used for continuos inputs. For categorical inputs, we always use
                the mode. Defaults to np.median.
            q (Collection[float], optional): List of quantiles used for quantile binning.
                Defaults to (0, 0.25, 0.5, 0.75, 1).
            sort_by_len (bool): If True, sort the splits by length.
                (Applies to the train/test/valid and train/test splits). Defaults to True.
        """
        self._grouping_q = year_q
        super().__init__(
            ds=ds,
            shuffle=shuffle,
            random_state=random_state,
            sample_frac=sample_frac,
            stratification_col=stratification_col,
            center=center,
            q=q,
            sort_by_len=sort_by_len,
        )

    def _get_groups(self) -> Collection[int]:
        return quantile_binning(self._ds.get_years(range(len(self._ds))), self._grouping_q)


class KennardStoneSplitter(BaseSplitter):
    """Run the Kennard-Stone sampling algorithm [KennardStone]_.

    The algorithm selects samples with uniform converage.
    The initial samples are biased towards the boundaries of the dataset.
    Hence, it might be biased by outliers.

    This algorithm ensures a flat coverage of the dataset.
    It is also known as CADEX algorithm and has been later refined
    in the DUPLEX algorithm [Snee]_.

    .. warning::

        This splitter can be slow for large datasets as
        it requires us to perform distance matrices N times for a dataset
        with N structures.


    .. warning::

        Stratification is not supported for this splitter.


    .. warning::

        I couldn't find a good reference for the k-fold version of
        this algorihm.
    """

    def __init__(
        self,
        ds: AbstractStructureDataset,
        feature_names: List[str],
        shuffle: bool = True,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        sample_frac: Optional[float] = 1.0,
        scale: bool = True,
        centrality_measure: str = "mean",
        metric: Union[Callable, str] = "euclidean",
        ascending: bool = False,
    ) -> None:
        """Construct a KennardStoneSplitter.

        Args:
            ds (AbstractStructureDataset): A structure dataset.
                The :code:`BaseSplitter` only requires the length magic method to be implemented.
                However, other splitters might require additional methods.
            feature_names (List[str]): Names of features to consider.
            shuffle (bool): If True, perform a shuffled split.
                Defaults to True.
            random_state (Union[int, np.random.RandomState], optional):
                Random state for the shuffling. Defaults to None.
            sample_frac (float, optional):
                This can be used for downsampling. It will randomly select a subset of
                indices from all indices *before* splittings. For instance :code:`sample_frac=0.8`
                will randomly select 80% of the indices before splitting.
                Defaults to 1.0.
            scale (bool): If True, apply z-score normalization
                prior to running the sampling. Defaults to True.
            centrality_measure (str): The first sample is selected to be
                maximally distanct from this value. It can be one of "mean", "median",
                "random". In case of "random" we simply select a random point.
                In the case of "mean" and "median" the initial point is maximally distanct
                from the mean and median of the feature matrix, respectively.
                Defaults to "mean".
            metric (Union[Callable, str]): The distance metric to use.
                If a string, the distance function can be ‘braycurtis’, ‘canberra’, ‘chebyshev’,
                ‘cityblock’, ‘correlation’, ‘cosine’, ‘dice’, ‘euclidean’, ‘hamming’, ‘jaccard’,
                ‘jensenshannon’, ‘kulsinski’, ‘kulczynski1’, ‘mahalanobis’, ‘matching’, ‘minkowski’,
                ‘rogerstanimoto’, ‘russellrao’, ‘seuclidean’, ‘sokalmichener’, ‘sokalsneath’, ‘sqeuclidean’,
                ‘yule’. Defaults to "euclidean".
            ascending (bool): If True, sort samples in asceding distance to the center.
                That is, the first samples (maximally distant to center) would be sampled last.
                Defaults to False.
        """
        self.feature_names = feature_names
        self.scale = scale
        self.centrality_measure = centrality_measure
        self.metric = metric
        self.ascending = ascending
        self._sorted_indices = None
        super().__init__(
            ds=ds,
            shuffle=shuffle,
            random_state=random_state,
            sample_frac=sample_frac,
            center=None,
            stratification_col=None,
            q=None,
        )

    def get_sorted_indices(self, ds: AbstractStructureDataset) -> Collection[int]:
        """Return a list of indices, sorted by similarity using the Kennard-Stone algorithm.

        The first sample will be maximally distant from the center.

        Args:
            ds (AbstractStructureDataset): A mofdscribe AbstractStructureDataset

        Returns:
            Collection[int]: Sorted indices.
        """
        if self._sorted_indices is None:
            feats = ds._df[self.feature_names].values

            indices = kennard_stone_sampling(
                feats,
                scale=self.scale,
                centrality_measure=self.centrality_measure,
                metric=self.metric,
            )

            if self.ascending:
                indices = indices[::-1]

            self._sorted_indices = indices
        return self._sorted_indices

    def train_test_split(self, frac_train: float = 0.7) -> Tuple[Collection[int], Collection[int]]:
        num_train_points = int(frac_train * len(self._ds))

        if self._shuffle:
            return (
                np.random.permutation(self.get_sorted_indices(self._ds)[:num_train_points]),
                np.random.permutation(self.get_sorted_indices(self._ds)[num_train_points:]),
            )
        return (
            self.get_sorted_indices(self._ds)[:num_train_points],
            self.get_sorted_indices(self._ds)[num_train_points:],
        )

    def train_valid_test_split(
        self, frac_train: float = 0.7, frac_valid: float = 0.1
    ) -> Tuple[Collection[int], Collection[int], Collection[int]]:
        num_train_points = int(frac_train * len(self._ds))
        num_valid_points = int(frac_valid * len(self._ds))

        if self._shuffle:
            return (
                np.random.permutation(self.get_sorted_indices(self._ds)[:num_train_points]),
                np.random.permutation(
                    self.get_sorted_indices(self._ds)[
                        num_train_points : num_train_points + num_valid_points
                    ]
                ),
                np.random.permutation(
                    self.get_sorted_indices(self._ds)[num_train_points + num_valid_points :]
                ),
            )
        return (
            self.get_sorted_indices(self._ds)[:num_train_points],
            self.get_sorted_indices(self._ds)[
                num_train_points : num_train_points + num_valid_points
            ],
            self.get_sorted_indices(self._ds)[num_train_points + num_valid_points :],
        )

    def k_fold(self, k=5) -> Tuple[Collection[int], Collection[int]]:
        kf = KFold(n_splits=k, shuffle=False, random_state=self._random_state)
        for train_index, test_index in kf.split(self.get_sorted_indices(self._ds)):
            if self._shuffle:
                train_index = np.random.permutation(train_index)
                test_index = np.random.permutation(test_index)
            yield train_index, test_index


class ClusterSplitter(BaseSplitter):
    """Split the data into clusters and use the clusters as groups.

    The approach has been proposed on
    `Kaggle <https://www.kaggle.com/code/lucamassaron/are-you-doing-cross-validation-the-best-way/notebook>`_.
    In principle, we perform the following steps:

        1. Scale the data (optional).
        2. Perform PCA for de-correlation.
        3. Perform k-means clustering.
    """

    def __init__(
        self,
        ds: AbstractStructureDataset,
        feature_names: List[str],
        shuffle: bool = True,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        sample_frac: Optional[float] = 1.0,
        stratification_col: Optional[Union[str, np.typing.ArrayLike]] = None,
        center: callable = np.median,
        q: Collection[float] = (0, 0.25, 0.5, 0.75, 1),
        sort_by_len: bool = False,
        scaled: bool = True,
        n_pca_components: Optional[Union[int, str]] = "mle",
        n_clusters: int = 4,
        pca_kwargs: Optional[Dict[str, Any]] = None,
        kmeans_kwargs: Optional[Dict[str, Any]] = None,
    ):
        """Construct a ClusterSplitter.

        Args:
            ds (AbstractStructureDataset): A structure dataset.
                The :code:`BaseSplitter` only requires the length magic method to be implemented.
                However, other splitters might require additional methods.
            feature_names (List[str]): Names of features to consider.
                        shuffle (bool): If True, perform a shuffled split.
                Defaults to True.
            shuffle (bool): If True, perform a shuffled split.
                Defaults to True.
            random_state (Union[int, np.random.RandomState], optional):
                Random state for the shuffling. Defaults to None.
            sample_frac (float, optional):
                This can be used for downsampling. It will randomly select a subset of
                indices from all indices *before* splittings. For instance :code:`sample_frac=0.8`
                will randomly select 80% of the indices before splitting.
                Defaults to 1.0.
            stratification_col (Union[str, np.typing.ArrayLike], optional): Data used for stratification.
                If it is categorical (see :py:meth:`mofdscribe.splitters.utils.is_categorical`)
                then we directly use it for stratification. Otherwise, we use quantile binning.
                Defaults to None.
            center (callable): Aggregation function to compute a measure of centrality
                of all the points in a group such that this can then be used for stratification.
                This is only used for continuos inputs. For categorical inputs, we always use
                the mode. Defaults to np.median.
            q (Collection[float], optional): List of quantiles used for quantile binning.
                Defaults to (0, 0.25, 0.5, 0.75, 1]. Defaults to [0, 0.25, 0.5, 0.75, 1).
            sort_by_len (bool): If True, sort the splits by length.
                (Applies to the train/test/valid and train/test splits). Defaults to True.
            scaled (bool): If True, scale the data before clustering.
                Defaults to True.
            n_pca_components (Union[int, str]): Number of components to use for PCA.
                If "mle", use the number of components that maximizes the variance.
                Defaults to "mle".
            n_clusters (int): Number of clusters to use.
                Defaults to 4.
            random_state (int): Random seed.
                Defaults to 42.
            pca_kwargs (Dict[str, Any]): Keyword arguments to pass to PCA.
                Defaults to None.
            kmeans_kwargs (Dict[str, Any]): Keyword arguments to pass to k-means.
                Defaults to None.
        """
        self.feature_names = feature_names
        self.scaled = scaled
        self.n_pca_components = n_pca_components
        self.n_clusters = n_clusters
        self._random_state = random_state
        self._sorted_indices = None
        self.ascending = False
        self._pca_kwargs = pca_kwargs
        self._kmeans_kwargs = kmeans_kwargs
        super().__init__(
            ds=ds,
            shuffle=shuffle,
            random_state=random_state,
            sample_frac=sample_frac,
            stratification_col=stratification_col,
            center=center,
            q=q,
            sort_by_len=sort_by_len,
        )

    def _get_sorted_indices(
        self, ds: AbstractStructureDataset, shuffle: bool = True
    ) -> Collection[int]:
        if self._sorted_indices is None:
            feats = ds._df[self.feature_names].values

            clusters = pca_kmeans(
                feats,
                n_clusters=self.n_clusters,
                n_pca_components=self.n_pca_components,
                random_state=self._random_state,
                scaled=self.scaled,
                pca_kwargs=self._pca_kwargs,
                kmeans_kwargs=self._kmeans_kwargs,
            )
            random_numbers = np.arange(len(clusters))
            if shuffle:
                np.random.shuffle(random_numbers)

            t = [(v, i, random_numbers[i]) for i, v in enumerate(clusters)]
            t.sort(reverse=not self.ascending, key=lambda x: (x[0], x[2]))
            indices = [i for _, i, _ in t]
            self._sorted_indices = indices
        return self._sorted_indices

    def _get_groups(self) -> Collection[Union[int, str]]:
        si = self._get_sorted_indices(self._ds, self._shuffle)
        return np.array(si)


class ClusterStratifiedSplitter(BaseSplitter):
    """Split the data into clusters and stratify on those clusters

    The approach has been proposed on
    `Kaggle <https://www.kaggle.com/code/lucamassaron/are-you-doing-cross-validation-the-best-way/notebook>`_.
    In principle, we perform the following steps:

    1. Scale the data (optional).
    2. Perform PCA for de-correlation.
    3. Perform k-means clustering.
    """

    def __init__(
        self,
        ds: AbstractStructureDataset,
        feature_names: List[str],
        shuffle: bool = True,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        sample_frac: Optional[float] = 1.0,
        scaled: bool = True,
        n_pca_components: Optional[int] = "mle",
        n_clusters: int = 4,
        pca_kwargs: Optional[Dict[str, Any]] = None,
        kmeans_kwargs: Optional[Dict[str, Any]] = None,
    ):
        """Construct a ClusterStratifiedSplitter.

        Args:
            ds (AbstractStructureDataset): A structure dataset.
                The :code:`BaseSplitter` only requires the length magic method to be implemented.
                However, other splitters might require additional methods.
            feature_names (List[str]): Names of features to consider.
            shuffle (bool): If True, perform a shuffled split.
                Defaults to True.
            random_state (Union[int, np.random.RandomState], optional):
                Random state for the shuffling. Defaults to None.
            sample_frac (float, optional):
                This can be used for downsampling. It will randomly select a subset of
                indices from all indices *before* splittings. For instance :code:`sample_frac=0.8`
                will randomly select 80% of the indices before splitting.
                Defaults to 1.0.
            scaled (bool): If True, scale the data before clustering.
                Defaults to True.
            n_pca_components (int): Number of components to use for PCA.
                If "mle", use the number of components that maximizes the variance.
                Defaults to "mle".
            n_clusters (int): Number of clusters to use.
                Defaults to 4.
            random_state (int): Random seed.
                Defaults to 42.
            pca_kwargs (Dict[str, Any]): Keyword arguments to pass to PCA.
                Defaults to None.
            kmeans_kwargs (Dict[str, Any]): Keyword arguments to pass to k-means.
                Defaults to None.
        """
        self.feature_names = feature_names
        self.scaled = scaled
        self.n_pca_components = n_pca_components
        self.n_clusters = n_clusters
        self._random_state = random_state
        self._stratification_groups = None
        self.ascending = False
        self._pca_kwargs = pca_kwargs
        self._kmeans_kwargs = kmeans_kwargs
        super().__init__(
            ds=ds,
            shuffle=shuffle,
            random_state=random_state,
            sample_frac=sample_frac,
            stratification_col=None,
            center=None,
            q=None,
        )

    def _get_stratification_col(self) -> Collection[int]:
        if self._stratification_groups is None:
            feats = self._ds._df[self.feature_names].values

            clusters = pca_kmeans(
                feats,
                n_clusters=self.n_clusters,
                n_pca_components=self.n_pca_components,
                random_state=self._random_state,
                scaled=self.scaled,
                pca_kwargs=self._pca_kwargs,
                kmeans_kwargs=self._kmeans_kwargs,
            )

            self._stratification_groups = clusters

        return self._stratification_groups


class LOCOCV(BaseSplitter):
    """
    Leave-one-cluster-out cross-validation.

    The general idea has been discussed before, e.g. in [Kramer]_.
    Perhaps more widely used in the materials community is [Meredig]_.
    Here, we perform PCA, followed by k-means clustering.

    * Where k = 2 for a train/test split
    * Where k = 3 for a train/valid/test split
    * Where k = k for k-fold crossvalidation

    By default, we will sort outputs such that the cluster sizes are
    train >= test >= valid.
    """

    def __init__(
        self,
        ds: AbstractStructureDataset,
        feature_names: List[str],
        shuffle: bool = True,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
        sample_frac: Optional[float] = 1.0,
        scaled: bool = True,
        n_pca_components: Optional[int] = "mle",
        pca_kwargs: Optional[Dict[str, Any]] = None,
        kmeans_kwargs: Optional[Dict[str, Any]] = None,
    ):
        """Construct a LOCOCV.

        Args:
            ds (AbstractStructureDataset): A structure dataset.
                The :code:`BaseSplitter` only requires the length magic method to be implemented.
                However, other splitters might require additional methods.
            feature_names (List[str]): Names of features to consider.
            shuffle (bool): If True, perform a shuffled split.
                Defaults to True.
            random_state (Union[int, np.random.RandomState], optional):
                Random state for the shuffling. Defaults to None.
            sample_frac (float, optional):
                This can be used for downsampling. It will randomly select a subset of
                indices from all indices *before* splittings. For instance :code:`sample_frac=0.8`
                will randomly select 80% of the indices before splitting.
                Defaults to 1.0.
            scaled (bool): If True, scale the data before clustering.
                Defaults to True.
            n_pca_components (int): Number of components to use for PCA.
                If "mle", use the number of components that maximizes the variance.
                Defaults to "mle".
            random_state (int): Random seed.
                Defaults to 42.
            pca_kwargs (Dict[str, Any], optional): Additional keyword arguments for
                sklearn's :py:class:`sklearn.decomposition.PCA`. Defaults to None.
            kmeans_kwargs (Dict[str, Any], optional):  Additional keyword arguments for
                sklearn's :py:class:`sklearn.clustering.KMeans`. Defaults to None.
        """
        self.scaled = scaled
        self.n_pca_components = n_pca_components
        self._random_state = random_state
        self._pca_kwargs = pca_kwargs
        self._kmeans_kwargs = kmeans_kwargs
        self._stratification_groups = None
        self.ascending = False
        self.feature_names = feature_names
        super().__init__(
            ds=ds,
            shuffle=shuffle,
            random_state=random_state,
            sample_frac=sample_frac,
            center=None,
            stratification_col=None,
            q=None,
        )

    def train_test_split(
        self,
    ) -> Tuple[Collection[int], Collection[int]]:
        """Perform a train/test partition.

        Returns:
            Tuple[Collection[int], Collection[int]]: Train indices, test indices
        """
        groups = pca_kmeans(
            self._ds._df[self.feature_names].values,
            scaled=self.scaled,
            n_pca_components=self.n_pca_components,
            n_clusters=2,
            random_state=self._random_state,
            pca_kwargs=self._pca_kwargs,
            kmeans_kwargs=self._kmeans_kwargs,
        )

        first_group = np.where(groups == 0)[0]
        second_group = np.where(groups == 1)[0]

        if self._shuffle:
            np.random.shuffle(first_group)
            np.random.shuffle(second_group)

        # potential downsampling after shuffle
        first_group = first_group[: int(self._sample_frac * len(first_group))]
        second_group = second_group[: int(self._sample_frac * len(second_group))]

        if len(first_group) > len(second_group):
            return first_group, second_group

        return second_group, first_group

    def train_valid_test_split(
        self,
    ) -> Tuple[Collection[int], Collection[int], Collection[int]]:
        """Perform a train/valid/test partition.

        Returns:
            Tuple[Collection[int], Collection[int], Collection[int]]: Training, validation, test set.
        """
        groups = pca_kmeans(
            self._ds._df[self.feature_names].values,
            scaled=self.scaled,
            n_pca_components=self.n_pca_components,
            n_clusters=3,
            random_state=self._random_state,
            pca_kwargs=self._pca_kwargs,
            kmeans_kwargs=self._kmeans_kwargs,
        )

        first_group = np.where(groups == 0)[0]
        second_group = np.where(groups == 1)[0]
        third_group = np.where(groups == 2)[0]

        if self._shuffle:
            np.random.shuffle(first_group)
            np.random.shuffle(second_group)
            np.random.shuffle(third_group)

        # potential downsampling after shuffle
        first_group = first_group[: int(self._sample_frac * len(first_group))]
        second_group = second_group[: int(self._sample_frac * len(second_group))]
        third_group = third_group[: int(self._sample_frac * len(third_group))]

        groups_sorted_by_len = sorted(
            [first_group, second_group, third_group], key=len, reverse=True
        )
        return groups_sorted_by_len[0], groups_sorted_by_len[2], groups_sorted_by_len[1]

    def k_fold(self, k: int) -> Tuple[Collection[int], Collection[int]]:
        """Peform k-fold crossvalidation.

        Args:
            k (int): Number of folds. Defaults to 5.

        Yields:
            Iterator[Tuple[Collection[int], Collection[int]]]: Train indices, test indices.
        """
        groups = pca_kmeans(
            self._ds._df[self.feature_names].values,
            scaled=self.scaled,
            n_pca_components=self.n_pca_components,
            n_clusters=k,
            random_state=self._random_state,
            pca_kwargs=self._pca_kwargs,
            kmeans_kwargs=self._kmeans_kwargs,
        )

        for group in range(k):
            train = np.where(groups != group)[0]
            test = np.where(groups == group)[0]
            if self._shuffle:
                np.random.shuffle(train)
                np.random.shuffle(test)
            # potential downsampling after shuffle
            train = train[: int(self._sample_frac * len(train))]
            test = test[: int(self._sample_frac * len(test))]
            if len(train) > len(test):
                yield train, test
            else:
                yield test, train



# -*- coding: utf-8 -*-
"""Test the PH Vect featurizer."""
import pytest

from mofdscribe.featurizers.topology.ph_vect import PHVect

from ..helpers import is_jsonable


def test_ph_vect(hkust_structure, irmof_structure, hkust_la_structure):
    """Ensure we get the correct number of features and that they are different for different structures."""
    # should raise if not fitted
    with pytest.raises(ValueError):
        ph_vect = PHVect()
        ph_vect.featurize(hkust_structure)

    # should be able to fit and featurize
    ph_vect = PHVect(n_components=2, random_state=42)
    ph_vect.fit([hkust_structure, irmof_structure])

    # test fit_transform
    ph_vect = PHVect(n_components=2, random_state=42)
    feat = ph_vect.fit_transform([hkust_structure, irmof_structure])
    assert feat.shape == (2, 4 * 2 * 2)
    assert is_jsonable(dict(zip(ph_vect.feature_labels(), feat[0])))
    assert feat.ndim == 2

    feat = ph_vect.featurize(hkust_structure)
    assert feat.ndim == 1

    assert len(feat) == len(set(ph_vect.feature_labels()))

    assert len(feat) == 4 * 2 * 2

    # test that we can encode chemistry with varying atomic radii
    ph_vect = PHVect(
        n_components=2, atom_types=None, alpha_weight="atomic_radius_calculated", random_state=42
    )
    hkust_feats = ph_vect.fit_transform([hkust_structure])
    features_hkust_la = ph_vect.fit_transform([hkust_la_structure])
    assert hkust_feats.shape == features_hkust_la.shape
    assert (hkust_feats != features_hkust_la).any()

    # now, to be sure do not encode the same thing with the atomic radius
    ph_vect = PHVect(n_components=2, atom_types=None, alpha_weight=None, random_state=42)
    hkust_feats = ph_vect.fit_transform([hkust_structure])
    features_hkust_la = ph_vect.fit_transform([hkust_la_structure])
    assert hkust_feats.shape == features_hkust_la.shape
    assert hkust_feats == pytest.approx(features_hkust_la, rel=1e-2)



# -*- coding: utf-8 -*-
"""Test CoRE dataset."""
import numpy as np
import pytest
from pymatgen.core import IStructure

from mofdscribe.datasets.core_dataset import CoREDataset


@pytest.mark.xdist_group(name="core-ds")
def test_core():
    """Ensure we can instantiate the CoRE dataset as access a few key methods."""
    core = CoREDataset()
    assert isinstance(list(core.get_structures([1]))[0], IStructure)
    assert isinstance(core.get_labels([1]), np.ndarray)
    assert isinstance(core.get_years([1]), np.ndarray)
    assert len(core.get_years([1])) == 1
    assert len(core.get_years([1, 2, 4])) == 3
    assert isinstance(core.get_labels([1], ["outputs.pure_methane_kH"]), np.ndarray)

    # make sure we can get a subset of the dataset
    core_subset = core.get_subset([1, 2, 3, 8])
    assert isinstance(core_subset, CoREDataset)
    assert len(core_subset._df) == 4
    assert (
        core._df.iloc[[1, 2, 3, 8]]["info.basename"].values
        == core_subset._df["info.basename"].values
    ).all()
    assert list(core.get_structures([1]))[0] == list(core_subset.get_structures([0]))[0]



"""Capture some basic information about the computing environment.

Taken from https://gist.github.com/tahajahangir/5275848#file-pyinfo-py
"""

import platform
import sys


def section_packages():
    data = []
    try:
        import pkg_resources
    except ImportError:
        return None
    for pkg in pkg_resources.working_set:
        assert isinstance(pkg, pkg_resources.Distribution)
        data.append((pkg.project_name, pkg.version if pkg.has_version() else "[uknown]"))
    return dict(sorted(data, key=lambda a: a[0].lower()))


def section_system():
    data = []
    if hasattr(sys, "subversion"):
        data.append(("Python Subversion", ", ".join(x for x in sys.subversion if x)))

        data.append(
            (
                "OS Version",
                "%s %s (%s %s)"
                % (
                    platform.system(),
                    platform.release(),
                    platform.dist()[0].capitalize(),
                    platform.dist()[1],
                ),
            )
        )
    else:
        data.append(("OS Version", "%s %s" % (platform.system(), platform.release())))
    if hasattr(sys, "executable"):
        data.append(("Executable", sys.executable))
    data.append(("Build Date", platform.python_build()[1]))
    data.append(("Compiler", platform.python_compiler()))
    if hasattr(sys, "api_version"):
        data.append(("Python API", sys.api_version))
    return dict(data)


def get_watermark():
    return {
        "packages": section_packages(),
        "system": section_system(),
    }



# -*- coding: utf-8 -*-
import numpy as np
import pytest

from mofdscribe.featurizers.utils.aggregators import ARRAY_AGGREGATORS, MA_ARRAY_AGGREGATORS


def test_aggregators():
    test_array = [1, 2, 3, 4, 5, 6, 7]
    masked_test_array = np.ma.array(
        test_array + [8], mask=[False, False, False, False, False, False, False, True]
    )

    true_values = {
        "sum": 28,
        "avg": 4,
        "max": 7,
        "min": 1,
        "range": 6,
        "std": 2,
        "mean": 4,
        "median": 4,
        "harmean": 2.6997245179063363,
        "geomean": 3.3800151591412964,
        "mad": 2.0,
        "trimean": 4.0,
        "inf": 7,
        "manhattan": 28,
    }

    for aggregator_name, aggregator in ARRAY_AGGREGATORS.items():
        masked_aggregator = MA_ARRAY_AGGREGATORS[aggregator_name]
        assert aggregator(test_array) == pytest.approx(true_values[aggregator_name])
        if aggregator_name == "inf":
            assert masked_aggregator(masked_test_array) == pytest.approx(8)
        elif aggregator_name == "manhattan":
            assert masked_aggregator(masked_test_array) == pytest.approx(36)
        else:
            assert masked_aggregator(masked_test_array) == pytest.approx(
                true_values[aggregator_name]
            )



# -*- coding: utf-8 -*-
"""Featurizers using persistent homology."""
from .atom_centered_ph import AtomCenteredPH, AtomCenteredPHSite  # noqa: F401
from .ph_hist import PHHist  # noqa: F401
from .ph_image import PHImage  # noqa: F401
from .ph_stats import PHStats  # noqa: F401
from .ph_vect import PHVect  # noqa: F401



# -*- coding: utf-8 -*-
"""Thermal Stability Dataset."""
import os
from typing import Collection, Optional, Tuple

import numpy as np
import pandas as pd
from loguru import logger

from mofdscribe.constants import MOFDSCRIBE_PYSTOW_MODULE
from mofdscribe.datasets.checks import check_all_file_exists, length_check
from mofdscribe.datasets.dataset import AbstractStructureDataset
from mofdscribe.datasets.utils import compress_dataset

__all__ = ["ThermalStabilityDataset"]


class ThermalStabilityDataset(AbstractStructureDataset):
    """Thermal stability for a subset of CoRE MOFs.

    Reproduced from [Nandy2022]_.
    Nandy et al. (2022) digitized traces from thermogravimetric analysis.
    The decomposition temperature they determined in this way is reported in
    `outputs.assigned_T_decomp`.

    To reduce the risk of data leakage, we  (by default) also only keep one representative
    structure for a "base refcode" (i.e. the first five letters of a refcode).
    For instance, the base refcode for IGAHED001 is IGAHED. Structures with same
    base refcode but different refcodes are often different refinements, or measurements
    at different temperatures and hence chemically quite similar. For instance,
    the base refcode `UMODEH` would appear 21 times, `KEDJAG` 17 times, and `UMOYOM` 17 times
    in the CoRE dataset used by Moosavi et al.
    Additionally, we (by default) only keep one structure per "structure hash"
    which is an approximate graph-isomoprhism check, assuming the VESTA bond thresholds
    for the derivation of the structure graph (e.g. the structure
    graph of ULOMAL occurs 59 in the CoRE database used by Moosavi et al.).

    The years refer to the publication dates of the paper crossreferenced
    in the CSD entry of the structure.

    The available labels are:

        * `outputs.assigned_T_decomp`: Decomposition temperature in Kelvin.

    References::
        .. [Nandy2022] `Nandy, A.; Terrones, G.; Arunachalam, N.; Duan, C.;
            Kastner, D. W.; Kulik, H. J.
            MOFSimplify, Machine Learning Models with Extracted Stability Data
            of Three Thousand Metal–Organic Frameworks.
            Scientific Data 2022, 9 (1). <https://doi.org/10.1038/s41597-022-01181-0>`_
    """

    _files = {
        "v0.0.1": {
            "df": "https://zenodo.org/record/7428485/files/data.json?download=1",
            "structures": "https://zenodo.org/record/7428485/files/structures.tar.gz?download=1",
            "expected_length": 2039,
        }
    }

    def __init__(
        self,
        version: str = "v0.0.1",
        drop_basename_duplicates: bool = True,
        drop_graph_duplicates: bool = True,
        subset: Optional[Collection[int]] = None,
        drop_nan: bool = False,
    ):
        """Construct an instance of the ThermalStabilityDataset.

        Args:
            version (str): version number to use.
                Defaults to "v0.0.1".
            drop_basename_duplicates (bool): If True, keep only one structure
                per CSD basename. Defaults to True.
            drop_graph_duplicates (bool): If True, keep only one structure
                per decorated graph hash. Defaults to True.
            subset (Collection[int], optional): indices of the structures to include.
                Defaults to None.
            drop_nan (bool): If True, drop rows with NaN values in features or hashes.
                Defaults to True.

        Raises:
            ValueError: If the provided version number is not available.
        """
        self._drop_basename_duplicates = drop_basename_duplicates
        self._drop_nan = drop_nan
        self._drop_graph_duplicates = drop_graph_duplicates
        if version not in self._files:
            raise ValueError(
                f"Version {version} not available. Available versions: {list(self._files.keys())}"
            )
        self.version = version

        self._structure_dir = MOFDSCRIBE_PYSTOW_MODULE.ensure_untar(
            "thermal-stability",
            self.version,
            name="structures.tar.gz",
            url=self._files[version]["structures"],
        )

        self._df = pd.DataFrame(
            MOFDSCRIBE_PYSTOW_MODULE.ensure_json(
                "thermal-stability", self.version, name="data.json", url=self._files[version]["df"]
            )
        ).reset_index(drop=True)

        compress_dataset(self._df)

        length_check(self._df, self._files[version]["expected_length"])

        if drop_basename_duplicates:
            old_len = len(self._df)
            self._df = self._df.drop_duplicates(subset=["info.basename"])
            logger.debug(
                f"Dropped {old_len - len(self._df)} duplicate basenames. New length {len(self._df)}"
            )
        if drop_graph_duplicates:
            old_len = len(self._df)
            self._df = self._df.drop_duplicates(subset=["info.decorated_graph_hash"])
            logger.debug(
                f"Dropped {old_len - len(self._df)} duplicate graphs. New length {len(self._df)}"
            )
        self._df = self._df.reset_index(drop=True)
        if drop_nan:
            self._df.dropna(
                subset=[c for c in self._df.columns if c.startswith("features.")]
                + [c for c in self._df.columns if c.startswith("info.")],
                inplace=True,
            )
            self._df.reset_index(drop=True, inplace=True)

        if subset is not None:
            self._df = self._df.iloc[subset]
            self._df = self._df.reset_index(drop=True)

        self._structures = [
            os.path.join(self._structure_dir, f + ".cif") for f in self._df["info.CoRE_name"]
        ]

        check_all_file_exists(self._structures)

        self._years = self._df["info.year"].values
        self._decorated_graph_hashes = self._df["info.decorated_graph_hash"].values
        self._undecorated_graph_hashes = self._df["info.undecorated_graph_hash"].values
        self._decorated_scaffold_hashes = self._df["info.decorated_scaffold_hash"].values
        self._undecorated_scaffold_hashes = self._df["info.undecorated_scaffold_hash"].values
        self._densities = self._df["info.density"].values
        self._labelnames = (c for c in self._df.columns if c.startswith("outputs."))
        self._featurenames = (c for c in self._df.columns if c.startswith("features."))
        self._infonames = (c for c in self._df.columns if c.startswith("info."))

    def get_subset(self, indices: Collection[int]) -> "AbstractStructureDataset":
        """Get a subset of the dataset.

        Args:
            indices (Collection[int]): indices of the structures to include.

        Returns:
            AbstractStructureDataset: a new dataset containing only the structures
                specified by the indices.
        """
        return ThermalStabilityDataset(
            version=self.version,
            drop_basename_duplicates=self._drop_basename_duplicates,
            drop_graph_duplicates=self._drop_graph_duplicates,
            subset=indices,
            drop_nan=self._drop_nan,
        )

    @property
    def available_info(self) -> Tuple[str]:
        return self._infonames

    @property
    def available_features(self) -> Tuple[str]:
        return self._featurenames

    @property
    def available_labels(self) -> Tuple[str]:
        return self._labelnames

    def get_labels(self, idx: Collection[int], labelnames: Collection[str] = None) -> np.ndarray:
        labelnames = labelnames if labelnames is not None else self._labelnames
        return self._df.iloc[idx][list(labelnames)].values

    @property
    def citations(self) -> Tuple[str]:
        return [
            "@article{Chung2019,"
            "doi = {10.1021/acs.jced.9b00835},"
            "url = {https://doi.org/10.1021/acs.jced.9b00835},"
            "year = {2019},"
            "month = nov,"
            "publisher = {American Chemical Society ({ACS})},"
            "volume = {64},"
            "number = {12},"
            "pages = {5985--5998},"
            "author = {Yongchul G. Chung and Emmanuel Haldoupis and Benjamin J. Bucior "
            "and Maciej Haranczyk and Seulchan Lee and Hongda Zhang and "
            "Konstantinos D. Vogiatzis and Marija Milisavljevic and Sanliang Ling "
            "and Jeffrey S. Camp and Ben Slater and J. Ilja Siepmann and "
            "David S. Sholl and Randall Q. Snurr},"
            "title = {Advances,  Updates,  and Analytics for the Computation-Ready, "
            "Experimental Metal{\textendash}Organic Framework Database: {CoRE} {MOF} 2019},"
            r"journal = {Journal of Chemical {\&}amp$\mathsemicolon$ Engineering Data}"
            "}",
            "@article{Chung2014,"
            "doi = {10.1021/cm502594j},"
            "url = {https://doi.org/10.1021/cm502594j},"
            "year = {2014},"
            "month = oct,"
            "publisher = {American Chemical Society ({ACS})},"
            "volume = {26},"
            "number = {21},"
            "pages = {6185--6192},"
            "author = {Yongchul G. Chung and Jeffrey Camp and "
            "Maciej Haranczyk and Benjamin J. Sikora and Wojciech Bury "
            "and Vaiva Krungleviciute and Taner Yildirim and Omar K. Farha "
            "and David S. Sholl and Randall Q. Snurr},"
            "title = {Computation-Ready,  Experimental Metal{\textendash}Organic Frameworks: "
            "A Tool To Enable High-Throughput Screening of Nanoporous Crystals},"
            "journal = {Chemistry of Materials}"
            "}",
            "@article{Nandy_2022,"
            "doi = {10.1038/s41597-022-01181-0},"
            "url = {https://doi.org/10.1038%2Fs41597-022-01181-0},"
            "year = 2022,"
            "month = {mar},"
            "publisher = {Springer Science and Business Media {LLC}},"
            "volume = {9},"
            "number = {1},"
            "author = {Aditya Nandy and Gianmarco Terrones and "
            "Naveen Arunachalam and Chenru Duan and David W. Kastner and Heather J. Kulik},"
            "title = {{MOFSimplify}, machine learning models with extracted stability data "
            "of three thousand metal{\textendash}organic frameworks},"
            "journal = {Sci Data}"
            "}",
        ]



# -*- coding: utf-8 -*-
"""Use RDkit featurizers on pymatgen molecules."""

from typing import Callable, Collection, List, Union

import numpy as np
from loguru import logger
from matminer.featurizers.base import BaseFeaturizer
from pymatgen.analysis.graphs import MoleculeGraph
from pymatgen.core import Molecule
from structuregraph_helpers.create import get_local_env_method

from .utils import create_rdkit_mol_from_mol_graph
from ..utils.extend import operates_on_imolecule, operates_on_molecule


@operates_on_molecule
@operates_on_imolecule
class RDKitAdaptor(BaseFeaturizer):
    """
    Use any featurizer that can operate on RDkit molecules on pymatgen molecules.

    For this, we convert the pymatgen molecule to an RDkit molecule,
    using the coordinates of the pymatgen molecule as the coordinates of
    a conformer.
    """

    def __init__(
        self,
        featurizer: Callable,
        feature_labels: Collection[str],
        local_env_strategy: str = "vesta",
        force_sanitize: bool = True,
    ) -> None:
        """Constuct a new RDKitAdaptor.

        Args:
            featurizer (Callable): Function that takes an RDKit molecule and returns
                some features (int, float, or list or array of them).
            feature_labels (Collection[str]): Names of features. Must be the same length as
                the number of features returned by the featurizer.
            local_env_strategy (str): If the `featurize` method is called with a `Molecule`
                object, this determines the local environment strategy to use to convert
                the molecule to a MoleculeGraph.
                Defaults to "vesta".
            force_sanitize (bool): If True, the RDKit molecule will be sanitized
        """
        self._featurizer = featurizer
        self._feature_labels = list(feature_labels)
        self._local_env_strategy = local_env_strategy
        self._force_sanitize = force_sanitize

    def __repr__(self) -> str:
        """Print string representation of the featurizer."""
        return "RDKitAdaptor(featurizer={}, feature_labels={})".format(
            self._featurizer, self._feature_labels
        )

    def _repr_mimebundle_(self, include=None, exclude=None):
        """Print string representation in Ipyython notebooks."""
        return self.__repr__()

    def feature_labels(self) -> List[str]:
        return self._feature_labels

    def featurize(self, molecule: Union[Molecule, MoleculeGraph]) -> np.ndarray:
        """
        Call the RDKit featurizer on the molecule.

        If the input molecule is a Molecule, we convert it to a MoleculeGraph
        using the local environment strategy specified in the constructor.

        Args:
            molecule: A pymatgen Molecule or MoleculeGraph object.

        Returns:
            A numpy array of features.
        """
        if isinstance(molecule, MoleculeGraph):
            molecule_graph = molecule
        else:
            molecule_graph = MoleculeGraph.with_local_env_strategy(
                molecule, get_local_env_method(self._local_env_strategy)
            )
        rdkit_mol = create_rdkit_mol_from_mol_graph(
            molecule_graph, force_sanitize=self._force_sanitize
        )
        feats = self._featurizer(rdkit_mol)
        if isinstance(feats, (list, tuple, np.ndarray)):
            return np.asarray(feats)
        elif isinstance(feats, (float, int)):
            return np.array([feats])
        else:
            logger.warning("Featurizer returned an unsupported type: {}".format(type(feats)))
            return feats

    def citations(self) -> List[str]:
        return [
            "@misc{https://doi.org/10.5281/zenodo.591637,"
            "doi = {10.5281/ZENODO.591637},"
            "url = {https://zenodo.org/record/591637},"
            "author = {Landrum,  Greg and Tosco,  Paolo and Kelley,"
            " Brian and {Ric} and {Sriniker} and {Gedeck} and Vianello,  "
            "Riccardo and {NadineSchneider} and Kawashima,"
            " Eisuke and Dalke,  Andrew and N,  Dan and Cosgrove,"
            " David and Jones,  Gareth and Cole,  Brian and Swain,"
            "  Matt and Turk,  Samo and {AlexanderSavelyev} and Vaucher,  Alain"
            " and Wójcikowski,  Maciej and {Ichiru Take} and Probst,  Daniel "
            "and Ujihara,  Kazuya and Scalfani,  Vincent F. and Godin,  Guillaume"
            " and Pahl,  Axel and {Francois Berenger} and {JLVarjo} "
            "and {Strets123} and {JP} and {DoliathGavid}},"
            "    title = {rdkit/rdkit: 2022_03_3 (Q1 2022) Release},"
            "    publisher = {Zenodo},"
            "    year = {2022},"
            "    copyright = {Open Access}"
            " }"
        ]

    def implementors(self) -> List[str]:
        return ["Kevin Maik Jablonka", "Greg Landrum and RDKit authors"]



import numpy as np
import pandas as pd
from matminer.featurizers.base import MultipleFeaturizer
from matminer.featurizers.structure import DensityFeatures

from mofdscribe.featurizers.base import MOFMultipleFeaturizer
from mofdscribe.featurizers.chemistry import RACS


def test_mofmultiplefeaturizer(hkust_structure, irmof_structure):
    """Test that the calls work. However, I currently do not know of a good
    way of testing if and how often get_primitive is called
    (other than looking at the logs)"""
    structures = [hkust_structure, irmof_structure]
    primitive_structures = [structure.get_primitive_structure() for structure in structures]
    featurizer = MOFMultipleFeaturizer([RACS(), DensityFeatures()], primitive=True)
    for featurizer_ in featurizer.featurizers:
        assert featurizer_.primitive is False

    # make sure the simple call works
    features = featurizer.featurize(irmof_structure)
    assert (
        len(features)
        == len(featurizer.feature_labels())
        == len(RACS().feature_labels()) + len(DensityFeatures().feature_labels())
    )

    featurizer_no_primitive = MOFMultipleFeaturizer([RACS(), DensityFeatures()], primitive=False)
    features_no_prim = featurizer_no_primitive.featurize(primitive_structures[-1])

    assert np.allclose(features, features_no_prim, rtol=1e-2, atol=1e-2, equal_nan=True)

    # now compare with using the "original" matminer MultipleFeaturizer
    featurizer_orig = MultipleFeaturizer([RACS(), DensityFeatures()])
    features_orig = featurizer_orig.featurize(primitive_structures[-1])
    assert np.allclose(features, features_orig, rtol=1e-2, atol=1e-2, equal_nan=True)

    # make sure the multiple calls work
    features_many_1 = featurizer.featurize_many(structures)
    features_many_1_labels = featurizer.feature_labels()
    assert features_many_1.ndim == 2

    featurizer = MOFMultipleFeaturizer(
        [RACS(), DensityFeatures()], iterate_over_entries=False, primitive=True
    )
    features_many_2 = featurizer.featurize_many(structures)
    features_many_2_labels = featurizer.feature_labels()
    assert features_many_2.ndim == 2

    features_many_1_df = pd.DataFrame(features_many_1, columns=features_many_1_labels)
    features_many_2_df = pd.DataFrame(features_many_2, columns=features_many_2_labels)

    # independent of the way we call the featurizer, the features should be the same
    values_1 = features_many_1_df.values
    values_2 = features_many_2_df[features_many_1_df.columns].values
    assert np.allclose(values_1, values_2, rtol=1e-2, atol=1e-2, equal_nan=True)



# -*- coding: utf-8 -*-
"""Featurizer that runs RASPA to calculate the Henry coefficient."""
import os
from glob import glob
from typing import List, Union

import numpy as np
from pymatgen.core import IStructure, Structure

from mofdscribe.featurizers.base import MOFBaseFeaturizer
from mofdscribe.featurizers.utils.extend import operates_on_istructure, operates_on_structure
from mofdscribe.featurizers.utils.raspa.base_parser import parse_base_output
from mofdscribe.featurizers.utils.raspa.resize_uc import resize_unit_cell
from mofdscribe.featurizers.utils.raspa.run_raspa import detect_raspa_dir, run_raspa

__all__ = ["Henry"]

WIDOM_INPUT_TEMPLATE = """SimulationType  MonteCarlo

Forcefield      Local
NumberOfCycles                {cycles}
PrintEvery {print_every}
NumberOfInitializationCycles  0

Framework 0
FrameworkName input
UnitCells {unit_cells}
UseChargesFromCIFFile yes

CutOff                        {cutoff}

ExternalTemperature {temp}


Component 0 MoleculeName             {molname}
            MoleculeDefinition       Local
            WidomProbability         1.0
            CreateNumberOfMolecules  0
"""


def parse_widom(directory: Union[str, os.PathLike]) -> dict:
    """Parse the widom output files in the given directory."""
    outputs = glob(os.path.join(directory, "Output", "System_0", "*.data"))
    if len(outputs) != 1:
        raise ValueError("Expected one output file, got {}".format(len(outputs)))

    parsed_output, _ = parse_base_output(outputs[0], system_name="System_0", ncomponents=1)

    component_results = list(parsed_output["components"].values())[0]

    return [
        component_results["henry_coefficient_average"],
        component_results["adsorption_energy_widom_average"],
    ], [
        component_results["henry_coefficient_dev"],
        component_results["adsorption_energy_widom_dev"],
    ]


@operates_on_structure
@operates_on_istructure
class Henry(MOFBaseFeaturizer):
    """Computes the Henry coefficient for a given molecule using the RASPA [1]_ program.

    While Henry coefficients are sometimes the targets of ML
    algorithms, they are sometimes also used as features. In fact, they are
    closely related to the energy histograms.
    """

    def __init__(
        self,
        raspa_dir: Union[str, os.PathLike, None] = None,
        cycles: int = 5_000,
        temperature: float = 300,
        cutoff: float = 12,
        mof_ff: str = "UFF",
        mol_ff: str = "TraPPE",
        mol_name: str = "CO2",
        tail_corrections: bool = True,
        mixing_rule: str = "Lorentz-Berthelot",
        shifted: bool = False,
        separate_interactions: bool = True,
        run_eqeq: bool = True,
        return_std: bool = False,
        primitive: bool = False,
    ):
        """Initialize the featurizer.

        Args:
            raspa_dir (Union[str, PathLike, None]): Path to the raspa
                directory (with lib, bin, share) subdirectories.
                If `None` we will look for the `RASPA_DIR` environment variable.
                Defaults to None.
            cycles (int): Number of simulation cycles.
                Defaults to 5_000.
            temperature (float): Simulation temperature in
                Kelvin. Defaults to 300.
            cutoff (float): Cutoff for simulation in Angstrom.
                Defaults to 12.
            mof_ff (str): Name of the forcefield used for the framework.
                Defaults to "UFF".
            mol_ff (str): Name of the forcefield used for the guest molecule.
                Defaults to "TraPPE".
            mol_name (str): Name of the guest molecule. Defaults to "CO2".
            tail_corrections (bool): If true, use analytical tail-correction
                for the contribution of the interaction potential after the
                cutoff. Defaults to True.
            mixing_rule (str): Mixing rule for framework and guest
                molecule force field. Available options are `Jorgenson` and
                `Lorentz-Berthelot`. Defaults to "Lorentz-Berthelot".
            shifted (bool): If true, shifts the potential to equal to zero at the
                cutoff. Defaults to False.
            separate_interactions (bool): If True use framework's force field
                for framework-molecule interactions.
                Defaults to True.
            run_eqeq (bool): If true, runs EqEq to compute charges.
                Defaults to True.
            return_std (bool): If true, return the standard deviations.
                Defaults to False.
            primitive (bool): If true, use the primitive unit cell.
                Defaults to False.

        Raises:
            ValueError: If the `RASPA_DIR` environment variable is not set.
        """
        self.raspa_dir = raspa_dir if raspa_dir else os.environ.get("RASPA_DIR")
        if self.raspa_dir is None:
            try:
                self.raspa_dir = detect_raspa_dir()
            except ValueError:
                raise ValueError(
                    "Please set the RASPA_DIR environment variable or provide the path for the class initialization."
                )
        self.cycles = cycles
        self.cutoff = cutoff
        self.mof_ff = mof_ff
        self.mol_ff = mol_ff
        self.mol_name = mol_name
        self.tail_corrections = tail_corrections
        self.mixing_rule = mixing_rule
        self.shifted = shifted
        self.separate_interactions = separate_interactions
        self.temperature = temperature
        self.run_eqeq = run_eqeq
        self.return_std = return_std
        super().__init__(primitive=primitive)

    def _featurize(self, s: Union[Structure, IStructure]) -> np.array:
        ff_molecules = {self.mol_name: self.mol_ff}

        parameters = {
            "ff_framework": self.mof_ff,
            "ff_molecules": ff_molecules,
            "shifted": self.shifted,
            "tail_corrections": self.tail_corrections,
            "mixing_rule": self.mixing_rule,
            "separate_interactions": self.separate_interactions,
        }
        replicas = resize_unit_cell(s, self.cutoff)
        ucells = f"{replicas[0]} {replicas[1]} {replicas[2]}"
        simulation_script = WIDOM_INPUT_TEMPLATE.format(
            cycles=self.cycles,
            unit_cells=ucells,
            cutoff=self.cutoff,
            print_every=self.cycles // 10,
            molname=self.mol_name,
            temp=self.temperature,
        )

        res, deviations = run_raspa(
            s,
            self.raspa_dir,
            simulation_script,
            parameters,
            parse_widom,
            self.run_eqeq,
        )

        if self.return_std:
            res.extend(deviations)
        return np.array(res)

    def feature_labels(self) -> List[str]:
        feat = [
            f"henry_coefficient_{self.mol_name}_{self.temperature}_mol/kg/Pa",
            f"heat_of_adsorption_{self.mol_name}_{self.temperature}_kJ/mol",
        ]
        if self.return_std:
            feat.extend(
                [
                    f"henry_coefficient_std_{self.mol_name}_{self.temperature}_mol/kg/Pa",
                    f"heat_of_adsorption_std_{self.mol_name}_{self.temperature}_kJ/mol",
                ]
            )
        return feat

    def implementors(self) -> List[str]:
        return ["Kevin Maik Jablonka", "Fergus Mcilwaine", "David Dubbeldam and RASPA authors"]

    def citations(self) -> List[str]:
        return [
            "@article{Dubbeldam2015,"
            "doi = {10.1080/08927022.2015.1010082},"
            "url = {https://doi.org/10.1080/08927022.2015.1010082},"
            "year = {2015},"
            "month = feb,"
            "publisher = {Informa {UK} Limited},"
            "volume = {42},"
            "number = {2},"
            "pages = {81--101},"
            r"author = {David Dubbeldam and Sof{'{\i}}a Calero and "
            "Donald E. Ellis and Randall Q. Snurr},"
            "title = {{RASPA}: molecular simulation software for adsorption "
            "and diffusion in flexible nanoporous materials},"
            "journal = {Molecular Simulation}"
            "}"
        ]



# -*- coding: utf-8 -*-
"""Test the Henry featurizer."""
from mofdscribe.featurizers.chemistry.henry import Henry

from ..helpers import is_jsonable


def test_henryhoa(hkust_structure, irmof_structure):
    """Make sure that the featurization works for typical MOFs and the number of features is as expected."""
    featurizer = Henry(cycles=10)
    features = featurizer.featurize(hkust_structure)
    labels = featurizer.feature_labels()
    assert len(features) == len(labels)
    features_irmof = featurizer.featurize(irmof_structure)
    assert sum(abs(features - features_irmof)) > 0
    assert is_jsonable(dict(zip(featurizer.feature_labels(), features)))
    assert features_irmof.ndim == 1

    featurizer = Henry(cycles=500, return_std=True)
    features = featurizer.featurize(hkust_structure)
    assert len(features) == len(featurizer.feature_labels()) == 4

    # make sure we indeed use pyeqeq
    featurizer = Henry(cycles=500, return_std=True, run_eqeq=False)
    features_no_charge = featurizer.featurize(hkust_structure)
    assert features_no_charge[0] < features[0]



# -*- coding: utf-8 -*-
"""Implement some shape featurizers from RDKit using the RDKitAdaptor."""
from typing import List

from rdkit.Chem.Descriptors3D import NPR1 as NPR1_RDKIT
from rdkit.Chem.Descriptors3D import NPR2 as NPR2_RDKIT
from rdkit.Chem.Descriptors3D import PMI1 as PMI1_RDKIT
from rdkit.Chem.Descriptors3D import PMI2 as PMI2_RDKIT
from rdkit.Chem.Descriptors3D import PMI3 as PMI3_RDKIT
from rdkit.Chem.Descriptors3D import Asphericity as Asphericity_RDKIT
from rdkit.Chem.Descriptors3D import Eccentricity as Eccentricity_RDKIT
from rdkit.Chem.Descriptors3D import InertialShapeFactor as InertialShapeFactor_RDKIT
from rdkit.Chem.Descriptors3D import RadiusOfGyration as RadiusOfGyration_RDKIT
from rdkit.Chem.Descriptors3D import SpherocityIndex as SpherocityIndex_RDKIT

from .rdkitadaptor import RDKitAdaptor


def _rod_likeness(mol):
    """Compute the ROD-likeness of a molecule."""
    return NPR2_RDKIT(mol) - NPR1_RDKIT(mol)


def _disk_likeness(mol):
    """Compute the disk-likeness of a molecule."""
    return 2 - 2 * NPR2_RDKIT(mol)


def _sphericity(mol):
    """Compute the sphericity of a molecule."""
    return NPR1_RDKIT(mol) + NPR2_RDKIT(mol) - 1


class Asphericity(RDKitAdaptor):
    """Featurizer for the RDKit Asphericity descriptor."""

    def __init__(self):
        """Construct a new Asphericity featurizer."""
        super().__init__(Asphericity_RDKIT, ["asphericity"])

    def citations(self) -> List[str]:
        return self.super().citations() + [
            "@incollection{Todeschini2008,"
            "doi = {10.1002/9783527618279.ch37},"
            "url = {https://doi.org/10.1002/9783527618279.ch37},"
            "year = {2008},"
            "month = may,"
            "publisher = {Wiley-{VCH} Verlag {GmbH}},"
            "pages = {1004--1033},"
            "author = {Roberto Todeschini and Viviana Consonni},"
            "title = {Descriptors from Molecular Geometry},"
            "booktitle = {Handbook of Chemoinformatics}"
            "}"
        ]


class Eccentricity(RDKitAdaptor):
    """Featurizer for the RDKit Eccentricity descriptor."""

    def __init__(self):
        """Construct a new Eccentricity featurizer."""
        super().__init__(Eccentricity_RDKIT, ["eccentricity"])

    def citations(self) -> List[str]:
        return self.super().citations() + [
            "@incollection{Todeschini2008,"
            "doi = {10.1002/9783527618279.ch37},"
            "url = {https://doi.org/10.1002/9783527618279.ch37},"
            "year = {2008},"
            "month = may,"
            "publisher = {Wiley-{VCH} Verlag {GmbH}},"
            "pages = {1004--1033},"
            "author = {Roberto Todeschini and Viviana Consonni},"
            "title = {Descriptors from Molecular Geometry},"
            "booktitle = {Handbook of Chemoinformatics}"
            "}"
        ]


class InertialShapeFactor(RDKitAdaptor):
    """Featurizer for the RDKit InertialShapeFactor descriptor."""

    def __init__(self):
        """Construct a new InertialShapeFactor featurizer."""
        super().__init__(InertialShapeFactor_RDKIT, ["inertial_shape_factor"])

    def citations(self) -> List[str]:
        return self.super().citations() + [
            "@incollection{Todeschini2008,"
            "doi = {10.1002/9783527618279.ch37},"
            "url = {https://doi.org/10.1002/9783527618279.ch37},"
            "year = {2008},"
            "month = may,"
            "publisher = {Wiley-{VCH} Verlag {GmbH}},"
            "pages = {1004--1033},"
            "author = {Roberto Todeschini and Viviana Consonni},"
            "title = {Descriptors from Molecular Geometry},"
            "booktitle = {Handbook of Chemoinformatics}"
            "}"
        ]


class NPR1(RDKitAdaptor):
    """Featurizer for the RDKit NPR1 descriptor."""

    def __init__(self):
        """Construct a new NPR1 featurizer."""
        super().__init__(NPR1_RDKIT, ["npr1"])

    def citations(self) -> List[str]:
        return self.super().citations() + [
            "@article{Sauer2003"
            "doi = {10.1021/ci025599w}"
            "url = {https://doi.org/10.1021/ci025599w}"
            "year = {2003}"
            "month = mar"
            "publisher = {American Chemical Society ({ACS})}"
            "volume = {43}"
            "number = {3}"
            "pages = {987--1003}"
            "author = {Wolfgang H. B. Sauer and Matthias K. Schwarz}"
            "title = {Molecular Shape Diversity of Combinatorial Libraries:"
            " A Prerequisite for Broad Bioactivity}"
            "journal = {Journal of Chemical Information and Computer Sciences"
            ""
        ]


class NPR2(RDKitAdaptor):
    """Featurizer for the RDKit NPR2 descriptor."""

    def __init__(self):
        """Construct a new NPR2 featurizer."""
        super().__init__(NPR2_RDKIT, ["npr2"])

    def citations(self) -> List[str]:
        return self.super().citations() + [
            "@article{Sauer2003"
            "doi = {10.1021/ci025599w}"
            "url = {https://doi.org/10.1021/ci025599w}"
            "year = {2003}"
            "month = mar"
            "publisher = {American Chemical Society ({ACS})}"
            "volume = {43}"
            "number = {3}"
            "pages = {987--1003}"
            "author = {Wolfgang H. B. Sauer and Matthias K. Schwarz}"
            "title = {Molecular Shape Diversity of Combinatorial Libraries:"
            " A Prerequisite for Broad Bioactivity}"
            "journal = {Journal of Chemical Information and Computer Sciences"
            ""
        ]


class PMI1(RDKitAdaptor):
    """Featurizer for the RDKit PMI1 descriptor."""

    def __init__(self):
        """Construct a new PMI1 featurizer."""
        super().__init__(PMI1_RDKIT, ["pmi1"])


class PMI2(RDKitAdaptor):
    """Featurizer for the RDKit PMI2 descriptor."""

    def __init__(self):
        """Construct a new PMI2 featurizer."""
        super().__init__(PMI2_RDKIT, ["pmi2"])


class PMI3(RDKitAdaptor):
    """Featurizer for the RDKit PMI3 descriptor."""

    def __init__(self):
        """Construct a new PMI3 featurizer."""
        super().__init__(PMI3_RDKIT, ["pmi3"])


class RadiusOfGyration(RDKitAdaptor):
    """Featurizer for the RDKit RadiusOfGyration descriptor."""

    def __init__(self):
        """Construct a new RadiusOfGyration featurizer."""
        super().__init__(RadiusOfGyration_RDKIT, ["radius_of_gyration"])

    def citations(self) -> List[str]:
        return self.super().citations() + [
            "@incollection{Todeschini2008,"
            "doi = {10.1002/9783527618279.ch37},"
            "url = {https://doi.org/10.1002/9783527618279.ch37},"
            "year = {2008},"
            "month = may,"
            "publisher = {Wiley-{VCH} Verlag {GmbH}},"
            "pages = {1004--1033},"
            "author = {Roberto Todeschini and Viviana Consonni},"
            "title = {Descriptors from Molecular Geometry},"
            "booktitle = {Handbook of Chemoinformatics}"
            "}"
        ]


class SpherocityIndex(RDKitAdaptor):
    """Featurizer for the RDKit Spherocity Index descriptor."""

    def __init__(self):
        """Construct a new SpherocityIndex featurizer."""
        super().__init__(SpherocityIndex_RDKIT, ["spherocity"])

    def citations(self) -> List[str]:
        return self.super().citations() + [
            "@incollection{Todeschini2008,"
            "doi = {10.1002/9783527618279.ch37},"
            "url = {https://doi.org/10.1002/9783527618279.ch37},"
            "year = {2008},"
            "month = may,"
            "publisher = {Wiley-{VCH} Verlag {GmbH}},"
            "pages = {1004--1033},"
            "author = {Roberto Todeschini and Viviana Consonni},"
            "title = {Descriptors from Molecular Geometry},"
            "booktitle = {Handbook of Chemoinformatics}"
            "}"
        ]


class RodLikeness(RDKitAdaptor):
    """Featurizer for the RDKit Rod Likeness descriptor.

    This descriptor is computed as NPR2 - NPR1.
    """

    def __init__(self):
        """Construct a new RodLikeness featurizer."""
        super().__init__(_rod_likeness, ["rod_likeness"])

    def citations(self) -> List[str]:
        return self.super().citations() + [
            "@article{Wirth2013,"
            "doi = {10.1007/s10822-013-9659-1},"
            "url = {https://doi.org/10.1007/s10822-013-9659-1},"
            "year = {2013},"
            "month = jun,"
            "publisher = {Springer Science and Business Media {LLC}},"
            "volume = {27},"
            "number = {6},"
            "pages = {511--524},"
            "author = {Matthias Wirth and Andrea Volkamer and Vincent Zoete"
            " and Friedrich Rippmann and Olivier Michielin and Matthias Rarey and Wolfgang H. B. Sauer},"
            "title = {Protein pocket and ligand shape comparison and its application in virtual screening},"
            "journal = {Journal of Computer-Aided Molecular Design}"
            "}"
        ]


class DiskLikeness(RDKitAdaptor):
    """Featurizer for the RDKit Disk Likeness descriptor.

    This descriptor is computed as 2 - 2 * NPR2
    """

    def __init__(self):
        """Construct a new DiskLikeness featurizer."""
        super().__init__(_disk_likeness, ["disk_likeness"])

    def citations(self) -> List[str]:
        return self.super().citations() + [
            "@article{Wirth2013,"
            "doi = {10.1007/s10822-013-9659-1},"
            "url = {https://doi.org/10.1007/s10822-013-9659-1},"
            "year = {2013},"
            "month = jun,"
            "publisher = {Springer Science and Business Media {LLC}},"
            "volume = {27},"
            "number = {6},"
            "pages = {511--524},"
            "author = {Matthias Wirth and Andrea Volkamer and Vincent Zoete"
            " and Friedrich Rippmann and Olivier Michielin and Matthias Rarey and Wolfgang H. B. Sauer},"
            "title = {Protein pocket and ligand shape comparison and its application in virtual screening},"
            "journal = {Journal of Computer-Aided Molecular Design}"
            "}"
        ]


class SphereLikeness(RDKitAdaptor):
    """Featurizer for the RDKit Sphere Likeness descriptor.

    This descriptor is computed as NPR1+NPR2-1.
    """

    def __init__(self):
        """Construct a new SphereLikeness featurizer."""
        super().__init__(_sphericity, ["sphere_likeness"])

    def citations(self) -> List[str]:
        return self.super().citations() + [
            "@article{Wirth2013,"
            "doi = {10.1007/s10822-013-9659-1},"
            "url = {https://doi.org/10.1007/s10822-013-9659-1},"
            "year = {2013},"
            "month = jun,"
            "publisher = {Springer Science and Business Media {LLC}},"
            "volume = {27},"
            "number = {6},"
            "pages = {511--524},"
            "author = {Matthias Wirth and Andrea Volkamer and Vincent Zoete"
            " and Friedrich Rippmann and Olivier Michielin and Matthias Rarey and Wolfgang H. B. Sauer},"
            "title = {Protein pocket and ligand shape comparison and its application in virtual screening},"
            "journal = {Journal of Computer-Aided Molecular Design}"
            "}"
        ]
